import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
# will ultimately allow to load several inputs/outputs models
from qtpy.QtWidgets import QWidget, QApplication, QGridLayout, QScrollArea, QVBoxLayout, QGroupBox, QPushButton
from epyseg.gui.open import OpenFileOrFolderWidget
import sys
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

# I think this stuff is ready to be overridden by the other stuff
# could add a popping preview otherwise...

# for what I wanna do I really need this stuff to implement all the functions of the OpenFileOrFolderWidget Class

class Multiple_inputs(QWidget):

    def __init__(self, nb_of_inputs=1, generic_label='input#', parent_window=None, finalize_text_change_method_overrider=None):
        super().__init__(parent=parent_window)
        self.nb_of_inputs = nb_of_inputs
        self.generic_label =generic_label
        self.finalize_text_change_method_overrider = finalize_text_change_method_overrider
        self.initUI()

    def initUI(self):
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.models_container = QWidget()

        self.scrollArea = QScrollArea()
        self.scrollArea.setWidgetResizable(True)

        # self.scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.models = []
        # self.model_weights = []
        # add window to it --> TODO
        # self.input_model = OpenFileOrFolderWidget(parent_window=self, is_file=True,
        #                                           extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5 *.json *.JSON *.model);;All Files (*)",
        #                                           tip_text='Drag and drop a single model file here')
        # self.input_weights = OpenFileOrFolderWidget(parent_window=self, label_text='Load weights',
        #                                             is_file=True,
        #                                             extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5);;All Files (*)",
        #                                             tip_text='Drag and drop a single weight file here')  # TODO shall i add *.model ???
        #

        self.models_container.layout = QVBoxLayout()
        # self.models_container.layout.setAlignment(Qt.AlignTop)
        # self.models_container.layout.setColumnStretch(0, 25)
        # self.models_container.layout.setColumnStretch(1, 75)
        # self.models_container.layout.setHorizontalSpacing(3)
        # self.models_container.layout.setVerticalSpacing(3)

        for iii in range(self.nb_of_inputs):
            self.add_new_model(iii)
            # self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()

        self.models_container.setLayout(self.models_container.layout)


        self.scrollArea.setWidget(self.models_container )

        # self.scrollArea.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        # self.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        # self.setFixedSize(self.size())

        layout.addWidget(self.scrollArea)
        # layout.addWidget(self.add_model_to_ensemble)
        # layout.addWidget(self.remove_model_to_ensemble)

        self.setLayout(layout)
        # self.setMinimumHeight(60)
        self.setFixedHeight(60)
        self.setMinimumWidth(400)




    #################################### in this block I reimplement all the methods of the Openfileorfolderwidget stuff

    # smartest would be to get the last modified model and update it --> see how this can be done...

    # weird command that returns the path to the first input that is gonna be used for the saving and also for the controls -> the idea is to mimmick a the behaviour of a single OpenFileOrFolderWidget object

    def get_items(self):
        if self.models is None:
            return None
        return [model.text() for model in self.models]


    # maybe all is ok this way
    def text(self):
        return self.models[0].text()

    # do override this to change the behaviour of inputs...
    # dangerous to override --> do it in a cleaner way
    # def finalize_text_change(self):
        # print('in', self.path.text())
        # pass

    def __get_last_modified_model(self):
        last_modified_time = 0
        last_modified_item = None
        for model in self.models:
            last_mod =model.get_time_of_last_change()
            if last_mod is not None:
                if last_mod>last_modified_time:
                    last_modified_time = last_mod
                    last_modified_item = model
        return last_modified_item


    def set_icon_ok(self, ok):
        model = self.__get_last_modified_model()
        if model is not None:
            model.set_icon_ok(ok)


        # self.models[0].set_icon_ok(ok)
        # if not self.show_ok_or_not_icon:
        #     return
        # if ok:
        #     self.ok_or_not_ico.setPixmap(self.ok_ico)
        # else:
        #     self.ok_or_not_ico.setPixmap(self.not_ok_ico)

    def set_size(self, size):
        model = self.__get_last_modified_model()
        # self.size_label.setText(size)
        if model is not None:
        # self.models[0].size_label.setText(size)
            model.size_label.setText(size)

    def get_list_using_glob(self):
        model = self.__get_last_modified_model()
        # self.size_label.setText(size)
        if model is not None:
            # self.models[0].size_label.setText(size)
            return model.get_list_using_glob()



    # def get_list_using_glob(self):
    #     pass

    #################################### end of the block I reimplement all the methods of the Openfileorfolderwidget stuff

    # this will be used to define the nb of inputs
    def set_nb_of_items(self, nb_items=1):
        # reset inputs
        self.models = []
        # empty layout then add new inputs
        try:
            for i in reversed(range(self.models_container.layout.count())):
                self.models_container.layout.itemAt(i).widget().setParent(None)
        except:
            pass
        for iii in range(nb_items):
            self.add_new_model(count=iii)

    def add_new_model(self, count=None):
        # self.groupBox_model = QGroupBox('Model '+str(len(self.models)), objectName='Model_'+str(len(self.models)))
        # self.groupBox_model.setEnabled(True)
        # groupBox layout
        self.groupBox_model_layout = QVBoxLayout()
        # self.groupBox_model_layout.setAlignment(Qt.AlignTop)
        # self.groupBox_model_layout.setColumnStretch(0, 25)
        # self.groupBox_model_layout.setColumnStretch(1, 25)
        # self.groupBox_model_layout.setColumnStretch(2, 50)
        # self.groupBox_model_layout.setColumnStretch(3, 2)
        # self.groupBox_model_layout.setHorizontalSpacing(3)
        # self.groupBox_model_layout.setVerticalSpacing(3)

        # add a groupbox per model

        # TODO maybe allow this to be stored and or reupdated --> not so easy would need to design specific stuff
        #TODO should I allow it to be saved in ini, most likely not...

        # add them the main checking function
        # difficulty is how to show preview of all models at once ???? --> Not so easy todo... --> do this later but ok for now


        # pb is that I miss all the features to update here...
        # shall I add the image on the side of the input maybe ??? --> maybe there is an idea there???
        # or find a trick

        # qqsdqdqsdqsd

        # do override the default method with the current one

        # EN FAIT C4EST VRAIMENT DANGEREUX --> FAIRE 9A PROPREMENT... --> DEFINE AN OVERRIDING METHOD THAT IS CALLED INSTEAD OF THE MAIN METHOD IF NOT NONE --> SUPER EASY AND CLEAN



        self.input_model = OpenFileOrFolderWidget(parent_window=self, add_timer_to_changetext=True,
                                                        show_ok_or_not_icon=True,  # label_text=label_input,
                                                        show_size=True,
                                                        tip_text='Drag and drop a single file or folder here',
                                                        objectName='open_input_button',
                                                        finalize_text_change_method_overrider=self.finalize_text_change_method_overrider
                                                        ) #objectName +'open_input_button'
        # self.input_weights = OpenFileOrFolderWidget(parent_window=self, label_text='Load weights (otpional)',
        #                                             is_file=True,
        #                                             extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5);;All Files (*)",
        #                                             tip_text='Drag and drop a single weight file here')  # TODO shall i add *.model ???


        self.models.append(self.input_model)
        # self.model_weights.append(self.input_weights)

        # add it to the layout too

        # self.groupBox_model_layout.addWidget(self.input_model)
        # self.groupBox_model_layout.addWidget(self.input_weights)

        # self.groupBox_model.setLayout(self.groupBox_model_layout)



        self.models_container.layout.addWidget(self.input_model)


    def remove_last_model(self):
        # for i in range(self.models_container.layout.count()):
        #     print(i)
        # for i in reversed(range(models_container.count())):
        try:
            self.models_container.layout.itemAt( self.models_container.layout.count()-1).widget().setParent(None)
        except:
            pass
        # self.models_container.layout.itemAt(self.models_container.layout.count()-1).widget().setParent(None)

        if self.models:
            self.models.pop()

    def get_models(self):
        # loop of other model lists
        # else return path to models and weights
        # models = {}
        # for iii,model in enumerate(self.models):
        #     if model.text() is not None:
        #         models_and_weights[iii]=[model.text(), self.model_weights[iii].text()]
        # if not models_and_weights:
        #     logger.error('No model loaded --> nothing can be done, sorry...')
        #     return None
        models = []
        for iii, model in enumerate(self.models):
            models.append(model.text())
        return models


if __name__ == '__main__':
    # just for a test

    def printous():
        print('heeloow')

    app = QApplication(sys.argv)
    ex = Multiple_inputs(nb_of_inputs=3, finalize_text_change_method_overrider=printous)
    ex.show()
    app.exec_()

    # that seems to work
    print(ex.get_models())



# TODO depending on selected mode activate or not stuff
# if output is not 0 --> then use min max as 0 1
# if none --> remove all parameters

import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtWidgets import QDialog, QDoubleSpinBox, QToolTip, QPushButton, QDialogButtonBox
from qtpy.QtWidgets import QApplication, QGridLayout
from qtpy.QtWidgets import QSpinBox, QComboBox, QVBoxLayout, QLabel, QCheckBox, QGroupBox
from qtpy.QtCore import Qt, QPoint
from qtpy import QtWidgets, QtCore
import sys
from epyseg.deeplearning.docs.doc2html import markdown_file_to_html
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()


class PostProcessGUI(QDialog):

    def __init__(self, parent_window=None, _is_dialog=False):
        super().__init__(parent=parent_window)
        self._is_dialog = _is_dialog
        self.initUI()

    def initUI(self):
        input_v_layout = QVBoxLayout()
        input_v_layout.setAlignment(Qt.AlignTop)
        input_v_layout.setContentsMargins(0, 0, 0, 0)
        # TODO add a set of parameters there for the post process
        self.groupBox_post_process = QGroupBox(
            'Refine segmentation/Create a binary mask', objectName='groupBox_post_process')
        self.groupBox_post_process.setCheckable(True)
        self.groupBox_post_process.setChecked(True)
        # self.groupBox_post_process.setEnabled(True)

        group_box_post_process_parameters_layout = QGridLayout()
        group_box_post_process_parameters_layout.setAlignment(Qt.AlignTop)
        group_box_post_process_parameters_layout.setHorizontalSpacing(3)
        group_box_post_process_parameters_layout.setVerticalSpacing(3)

        # do a radio dialog with all the stuff needed...
        # test all

        post_process_method_selection_label = QLabel('Post process method')  # (or bond score for pretrained model)
        post_process_method_selection_label.setStyleSheet("QLabel { color : red; }")

        self.post_process_method_selection = QComboBox(objectName='post_process_method_selection')
        self.post_process_method_selection.addItem('Default (Slow/robust) (EPySeg pre-trained model only!)')
        self.post_process_method_selection.addItem('Fast (May contain more errors) (EPySeg pre-trained model only!)')
        self.post_process_method_selection.addItem('Old method (Sometimes better than default) (EPySeg pre-trained model only!)')
        self.post_process_method_selection.addItem('Simply binarize output using threshold')
        self.post_process_method_selection.addItem('Keep first channel only')
        self.post_process_method_selection.addItem('None (Raw model output)')
        self.post_process_method_selection.currentTextChanged.connect(self._post_process_method_changed)

        group_box_post_process_parameters_layout.addWidget(post_process_method_selection_label, 0, 0, 1, 1)
        group_box_post_process_parameters_layout.addWidget(self.post_process_method_selection, 0, 1, 1, 3)

        # TODO --> always make this relative
        threshold_label = QLabel(
            'Threshold: (in case of over/under segmentation, please increase/decrease, respectively)')  # (or bond score for pretrained model)
        threshold_label.setStyleSheet("QLabel { color : red; }")
        self.threshold_bond_or_binarisation = QDoubleSpinBox(objectName='threshold_bond_or_binarisation')
        self.threshold_bond_or_binarisation.setSingleStep(0.01)
        self.threshold_bond_or_binarisation.setRange(0.01, 1)  # 100_000 makes no sense (oom) but anyway
        self.threshold_bond_or_binarisation.setValue(0.42)  # probably should be 1 to 3 depending on the tissue
        self.threshold_bond_or_binarisation.setEnabled(False)
        # threshold_hint = QLabel()  # (or bond score for pretrained model)

        self.autothreshold = QCheckBox("Auto",objectName='autothreshold')
        self.autothreshold.setChecked(True)
        self.autothreshold.stateChanged.connect(self._threshold_changed)

        group_box_post_process_parameters_layout.addWidget(threshold_label, 1, 0, 1, 2)
        group_box_post_process_parameters_layout.addWidget(self.threshold_bond_or_binarisation, 1, 2)
        group_box_post_process_parameters_layout.addWidget(self.autothreshold, 1, 3)
        # groupBox_post_process_parameters_layout.addWidget(threshold_hint, 0, 3)

        filter_by_size_label = QLabel('Further filter segmentation by size:')
        self.filter_by_cell_size_combo = QComboBox(objectName='filter_by_cell_size_combo')
        self.filter_by_cell_size_combo.addItem('None (quite often the best choice)')
        self.filter_by_cell_size_combo.addItem('Local median (slow/very good) divided by')
        self.filter_by_cell_size_combo.addItem('Cells below Average area (global) divided by')
        self.filter_by_cell_size_combo.addItem('Global median divided by')
        self.filter_by_cell_size_combo.addItem('Cells below size (in px)')

        # add a listener to model Architecture
        self.filter_by_cell_size_combo.currentTextChanged.connect(self._filter_changed)

        group_box_post_process_parameters_layout.addWidget(filter_by_size_label, 2, 0)
        group_box_post_process_parameters_layout.addWidget(self.filter_by_cell_size_combo, 2, 1, 1, 2)

        self.avg_area_division_or_size_spinbox = QSpinBox(objectName='avg_area_division_or_size_spinbox')
        self.avg_area_division_or_size_spinbox.setSingleStep(1)
        self.avg_area_division_or_size_spinbox.setRange(1, 10000000)  # 100_000 makes no sense (oom) but anyway
        self.avg_area_division_or_size_spinbox.setValue(2)  # probably should be 1 to 3 depending on the tissue
        self.avg_area_division_or_size_spinbox.setEnabled(False)
        group_box_post_process_parameters_layout.addWidget(self.avg_area_division_or_size_spinbox, 2, 3)

        self.prevent_exclusion_of_too_many_cells_together = QCheckBox('Do not exclude groups bigger than', objectName='prevent_exclusion_of_too_many_cells_together')
        self.prevent_exclusion_of_too_many_cells_together.setChecked(False)
        self.prevent_exclusion_of_too_many_cells_together.setEnabled(False)

        # max_nb_of_cells_to_be_excluded_together_label = QLabel('Group size')
        self.max_nb_of_cells_to_be_excluded_together_spinbox = QSpinBox(objectName='max_nb_of_cells_to_be_excluded_together_spinbox')
        self.max_nb_of_cells_to_be_excluded_together_spinbox.setSingleStep(1)
        self.max_nb_of_cells_to_be_excluded_together_spinbox.setRange(1, 10000000)  # max makes no sense
        self.max_nb_of_cells_to_be_excluded_together_spinbox.setValue(
            3)  # default should be 2 or 3 because seg is quite good so above makes no sense
        self.max_nb_of_cells_to_be_excluded_together_spinbox.setEnabled(False)
        cells_text_labels = QLabel('cells')

        self.restore_secure_cells = QCheckBox('Restore most likely cells',objectName='restore_secure_cells')
        self.restore_secure_cells.setChecked(False)
        self.restore_secure_cells.setEnabled(False)

        # help for post process
        # help_ico = QIcon.fromTheme('help-contents')
        self.help_button_postproc = QPushButton('?', None)
        bt_width = self.help_button_postproc.fontMetrics().boundingRect(self.help_button_postproc.text()).width() + 7
        self.help_button_postproc.setMaximumWidth(bt_width * 2)
        self.help_button_postproc.clicked.connect(self.show_tip)

        group_box_post_process_parameters_layout.addWidget(self.restore_secure_cells, 3, 0)
        group_box_post_process_parameters_layout.addWidget(self.prevent_exclusion_of_too_many_cells_together, 3, 1)
        group_box_post_process_parameters_layout.addWidget(self.max_nb_of_cells_to_be_excluded_together_spinbox, 3, 2)
        group_box_post_process_parameters_layout.addWidget(cells_text_labels, 3, 3)

        # TODO --> improve layout to make help button smaller
        group_box_post_process_parameters_layout.addWidget(self.help_button_postproc, 0, 5, 3, 1)

        self.groupBox_post_process.setLayout(group_box_post_process_parameters_layout)
        input_v_layout.addWidget(self.groupBox_post_process)
        self.setLayout(input_v_layout)

        if self._is_dialog:
            # OK and Cancel buttons
            self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                                            QtCore.Qt.Horizontal, self)
            self.buttons.accepted.connect(self.accept)
            self.buttons.rejected.connect(self.reject)
            self.layout().addWidget(self.buttons)

    def _threshold_changed(self):
        self.threshold_bond_or_binarisation.setEnabled(not self.autothreshold.isChecked())

    # self.post_process_method_selection.addItem('Default (Slow/robust)')
    # self.post_process_method_selection.addItem('Fast (May contain more errors)')
    # self.post_process_method_selection.addItem('Old method (Less constant than default but sometimes better)')
    # self.post_process_method_selection.addItem('Simply binarize output using threshold')
    # self.post_process_method_selection.addItem('None (Raw model output)')

    def _post_process_method_changed(self):
        text = self.post_process_method_selection.currentText().lower()
        if 'none' in text or 'first' in text:
            self.set_threshold_enabled(False)
            self.set_safety_parameters(False)
            self.set_filter_by_size_enabled(False)
        elif 'simply' in text:
            self.set_threshold_enabled(True)
            self.set_safety_parameters(False)
            self.set_filter_by_size_enabled(False)
        elif 'old' in text:
            self.set_threshold_enabled(False)
            self.set_safety_parameters(True)
            self.set_filter_by_size_enabled(True)
        else:
            self.set_threshold_enabled(True)
            self.set_safety_parameters(False)
            self.set_filter_by_size_enabled(True)

    def set_filter_by_size_enabled(self, bool):
        if bool is False:
            self.filter_by_cell_size_combo.setEnabled(False)
            self.avg_area_division_or_size_spinbox.setEnabled(False)
        else:
            self.filter_by_cell_size_combo.setEnabled(True)
            self.avg_area_division_or_size_spinbox.setEnabled(True)

    def set_threshold_enabled(self, bool):
        if bool is False:
            self.autothreshold.setEnabled(False)
            self.threshold_bond_or_binarisation.setEnabled(False)
        else:
            self.autothreshold.setEnabled(True)
            self._threshold_changed()

    def set_safety_parameters(self, bool):
        self._filter_changed()

    def show_tip(self):
        QToolTip.showText(self.sender().mapToGlobal(QPoint(30, 30)), markdown_file_to_html('refine_segmentation.md'))

    def isChecked(self):
        return self.groupBox_post_process.isChecked()

    def setChecked(self, bool):
        return self.groupBox_post_process.setChecked(bool)

    def _filter_changed(self):
        current_filter = self.filter_by_cell_size_combo.currentText().lower()
        current_mode = self.post_process_method_selection.currentText().lower()
        if 'one' in current_filter:
            self.avg_area_division_or_size_spinbox.setEnabled(False)
            self.max_nb_of_cells_to_be_excluded_together_spinbox.setEnabled(False)
            self.prevent_exclusion_of_too_many_cells_together.setEnabled(False)
            self.restore_secure_cells.setEnabled(False)
        else:
            self.avg_area_division_or_size_spinbox.setEnabled(True)
            self.max_nb_of_cells_to_be_excluded_together_spinbox.setEnabled(True)
            self.prevent_exclusion_of_too_many_cells_together.setEnabled(True)
            self.restore_secure_cells.setEnabled(True)
            if 'divided' in current_filter:
                self.avg_area_division_or_size_spinbox.setValue(2)
            else:
                self.avg_area_division_or_size_spinbox.setValue(300)
            if not 'old' in current_mode:
                self.max_nb_of_cells_to_be_excluded_together_spinbox.setEnabled(False)
                self.prevent_exclusion_of_too_many_cells_together.setEnabled(False)
                self.restore_secure_cells.setEnabled(False)

    def _get_post_process_filter(self):
        current_filter = self.filter_by_cell_size_combo.currentText().lower()
        if 'one' in current_filter or not self.filter_by_cell_size_combo.isEnabled():
            return None
        if 'size' in current_filter:
            return self.avg_area_division_or_size_spinbox.value()
        if 'verage' in current_filter:
            return 'avg'
        if 'local' in current_filter:
            return 'local'
        if 'global' in current_filter:
            return 'global median'

    def get_parameters_directly(self):
        '''Get the parameters for model training

            Returns
            -------
            dict
                containing post processing parameters

            '''

        self.post_process_parameters = {}
        post_proc_method = self.post_process_method_selection.currentText().lower()
        if 'none' in post_proc_method:
            self.post_process_parameters['post_process_algorithm'] = None
        else:
            self.post_process_parameters['post_process_algorithm'] = post_proc_method
        self.post_process_parameters['filter'] = self._get_post_process_filter()
        if self.threshold_bond_or_binarisation.isEnabled():
            self.post_process_parameters['threshold'] = self.threshold_bond_or_binarisation.value()
        if self.autothreshold.isEnabled() and self.autothreshold.isChecked():
            self.post_process_parameters[
                'threshold'] = None  # None means autothrehsold # maybe add more options some day
        if self.avg_area_division_or_size_spinbox.isEnabled():
            self.post_process_parameters['correction_factor'] = self.avg_area_division_or_size_spinbox.value()
        if self.restore_secure_cells.isEnabled():
            self.post_process_parameters['restore_safe_cells'] = self.restore_secure_cells.isChecked()
        if self.max_nb_of_cells_to_be_excluded_together_spinbox.isEnabled():
            self.post_process_parameters[
                'cutoff_cell_fusion'] = self.max_nb_of_cells_to_be_excluded_together_spinbox.value() if self.prevent_exclusion_of_too_many_cells_together.isChecked() else None

        if 'old' in self.post_process_method_selection.currentText().lower():
            # just for max use that --> maybe do this as an option some day
            self.post_process_parameters['hq_predictions'] = 'max'

        return self.post_process_parameters

    def get_parameters(self):
        return (self.get_parameters_directly())

    @staticmethod
    def getDataAndParameters(parent_window=None, _is_dialog=False):
        # get all the params for augmentation
        dialog = PostProcessGUI(parent_window=parent_window, _is_dialog=_is_dialog)
        result = dialog.exec_()
        parameters = dialog.get_parameters()
        return (parameters, result == QDialog.Accepted)


# now really try to get the parameters properly

if __name__ == '__main__':
    # just for a test
    app = QApplication(sys.argv)
    parameters, ok = PostProcessGUI.getDataAndParameters(parent_window=None)
    print(parameters, ok)
    sys.exit(0)

# TODO change default parameters depending on whether a pre-trained model is selected or not
# TODO allow retraining of the model --> just give it a try...



# MEGA TODO https://docs.python.org/3/library/multiprocessing.shared_memory.html --> can I block shared memory to prevent tf issues ???
# https://stackoverflow.com/questions/14124588/shared-memory-in-multiprocessing --> or maybe use the real true shared mem ???
# https://stackoverflow.com/questions/55277138/tensorflow-keras-multiprocessing-results-in-lack-of-gpu-memory --> mayeb this is my solution
# https://stackoverflow.com/questions/43641543/how-to-avoid-loading-a-parent-module-in-a-forked-process-with-pythons-multiproce
# https://stackoverflow.com/questions/42504669/keras-tensorflow-and-multiprocessing-in-python --> ok this is how I should do that then maybe --> TODO give it a try --> quite easy TODO

# https://github.com/yuanyuanli85/Keras-Multiple-Process-Prediction

# https://stackoverflow.com/questions/70263918/how-to-prevent-multiprocessing-from-inheriting-imports-and-globals --> maybe I need spawn --> can I get it to run ??? and if that does not work can I get it to default to something else
# can I close deep learning and tensorflow before MT processing --> maybe ???
# to reprodcue the error --> run deep learning first then run ta finish all and everything will be stuck forever

# do a small test to reproduce the MT error --> probably not so hard in fact



import numpy as np
from epyseg.tools.logger import TA_logger # logging
from epyseg.ta.utils.rps_tools import pavlidis, compute_distance_between_consecutive_points

logger = TA_logger()

# this would actually work for 2D and 3D area --> heightmap is optional but it makes no sense to run this with 2D coords and no heightmaps
# if a heightmap is provided the z coords of every point will be replaced by height map height
# nb if several triangles are provided, it is assumed they are all traingulation of the same object, hence their area is summed and the sum is returned
def compute_3D_surfacearea(triangles, heightmap=None, stop_on_2D=False, disable_2D_warning=False):
    """

    :param triangles: cell/object triangulation as input (as list or ndarray)
    :param heightmap: a heightmap for the cell
    :param stop_on_2D: break if 2D data is provided
    :return: the surface area of the triangulated object
    """
    if not isinstance(triangles, np.ndarray):
        triangles = np.asarray(triangles)

    if triangles.shape[-1] == 2 and heightmap is None:
        if not disable_2D_warning:
            logger.error(
            'Error, triangle coords are purely 2D and no heightmap is provided, hence it is not possible to compute 3D area. To continue, the software will assume all heights are the same (= 0)  (i.e. the image is really a pure 2D image)')  # assuming 2D maybe --> just stack

        if stop_on_2D:
            return

        flat_height = np.zeros((*triangles.shape[0:-1], (1)))
        triangles = np.append(triangles, flat_height, -1)

    # MEGA TODO NB SHOULD I MAKE Z THE FIRST DIMENSION ACTUALLY ????, probably does not change anything
    if heightmap is not None:
        if triangles.shape[-1] == 3:
            b = triangles[:, :, :-1]  # get rid of z before adding it again
        else:
            b=triangles
        search_coords = (b[..., 0].ravel(), b[..., 1].ravel())  # 'np.where' like coords
        heights = heightmap[search_coords]

        # print('heights', heights) # en fait c'est bon car des triangles sont out -> modifie l'aire

        heights = np.reshape(heights, (*b.shape[0:-1], (1)))
        triangles = np.append(b, heights, -1)  # add the height column to the coords

    # compute a, b and c bond lengths for 3D triangles
    a = ((triangles[:, 0, 0] - triangles[:, 1, 0]) ** 2 + (
            triangles[:, 0, 1] - triangles[:, 1, 1]) ** 2 +
         (triangles[:, 0, 2] - triangles[:, 1, 2]) ** 2) ** 0.5
    b = ((triangles[:, 1, 0] - triangles[:, 2, 0]) ** 2 + (
            triangles[:, 1, 1] - triangles[:, 2, 1]) ** 2 +
         (triangles[:, 1, 2] - triangles[:, 2, 2]) ** 2) ** 0.5
    c = ((triangles[:, 2, 0] - triangles[:, 0, 0]) ** 2 + (
            triangles[:, 2, 1] - triangles[:, 0, 1]) ** 2 +
         (triangles[:, 2, 2] - triangles[:, 0, 2]) ** 2) ** 0.5



    # compute the triangle area using Heron's formula
    # https://en.wikipedia.org/wiki/Triangle
    s = (a + b + c) / 2.0  # semi perimeter (half of the triangle's perimeter)
    areas = (s * (s - a) * (s - b) * (s - c)) ** 0.5

    return areas.sum()



# def compute3D_curved_distance(point1, point2, heightmap=None, three_points_mode=False):
#     if heightmap is None:
#         # cannot compute 3D dist
#         print('no valid height map --> cannot compute curved distance')
#         return distance_between_points(point1, point2)
#
#     pt1_y = int(point1[0])
#     pt1_x = int(point1[1])
#     # pt1_z = point1[2]
#     pt2_y = int(point2[0])
#     pt2_x = int(point2[1])
#     # pt2_z = point1[2]
#     rr, cc = draw.line(pt1_y, pt1_x,
#                        pt2_y, pt2_x)
#     # for every point I need get the height of it and compute the real 3D distance --> not hard in fact I think
#
#     dist3D = 0
#
#     # print(rr, cc,'-->')
#
#     if three_points_mode:
#         # dirty hack to speed up things but ok to find the stuff then really recompute the real distance --> should be ok
#         rr = [rr[0], rr[int(len(rr) / 2.)], rr[len(rr) - 1]]
#         cc = [cc[0], cc[int(len(cc) / 2.)], cc[len(cc) - 1]]
#
#     # print(rr,cc)
#     # compute distance of current point with prev one --> TODO
#     for iii in range(1, len(rr), 1):
#         y1 = rr[iii - 1]
#         x1 = cc[iii - 1]
#         z1 = heightmap[y1, x1]
#         y2 = rr[iii]
#         x2 = cc[iii]
#         z2 = heightmap[y2, x2]
#         # print(y1, x1, 'height', z1)
#         # print(y2, x2, 'height', z2)
#
#         # print(y1, x1, z1, y2, x2, z2)
#         # print('--> ', distance3D(y1, x1, z1, y2, x2, z2))
#         # NB POINTS NEED BE SORTED --> CHECK THAT THIS IS THE CASE --> EASY MAX DIST BETWEEN CONSECUTIVE 2D POINTS CAN ONLY BE SQRT 2
#
#         dist3D += distance3D(y1, x1, z1, y2, x2, z2)
#     # print(dist3D)
#     return dist3D

# TODO --> do that --> need sorted stuff
# if points are not sorted then need pavlidis sort them
# do pretty much the same as in the other
def perimeter_3D(points, heightmap,points_are_sorted=False,is_closed_contour=False):
    if not points_are_sorted:
        sorted_points =pavlidis(points, closed_contour=is_closed_contour)
    else:
        sorted_points = points

    if not isinstance(sorted_points, np.ndarray):
        sorted_points = np.asarray(sorted_points)

    # NOW NEED GET HEIGHT FOR ALL POINTS --> SHOULD BE DOABLE

    search_coords = (sorted_points[..., 0].ravel(), sorted_points[..., 1].ravel())  # 'np.where' like coords
    # print('search_coords',search_coords)
    heights = heightmap[search_coords]


    # print(sorted_points.shape)
    # print(heights.shape)
    # print(heights)
    heights=heights[..., np.newaxis]
    # sorted_points = np.append(sorted_points, heights, -1) # add the heights dimension to the sorted points
    sorted_points = np.append(heights,sorted_points, -1) # add the heights dimension to the sorted points
    # sorted_points = np.hstack((heights, sorted_points))  # add the heights dimension to the sorted points

    # print('sorted_points',sorted_points)
    consecutive_distance = compute_distance_between_consecutive_points(sorted_points)

    # print(consecutive_distance)



    perimeter3D = np.sum(consecutive_distance)


    # then compare to the pixel by pixel method of it and should be the same result in fact
    # then for every point sum the distance from one point to the next...



    # also find a way to order pixels and get the perimeter right
    # si > sqrt 2 alors need a connection --> put this as a section and try connect to first or last of every list --> maybe not so hard to implement and in the end all the points should be connected but try on a real example !!!
    # TODO
    # do it both for 2D and 3D
    # the ordering of the pixels is key --> really need do that
    # then I'll have a fairly good clone of TA that would be great and fully open source!!!
    # raise NotImplementedError


    return perimeter3D


if __name__ == '__main__':

        if False:
            perimeter_3D()
            import sys
            sys.exit(0)

        # it all seems to work but now just compute the area in a smarter way

        # triangles3D= [[[2.37749672, -0.99445809,0], [1.20416474, -0.9998551,1], [2.87830002, -0.99819343,2]],[[2.37749672, -0.99445809,0], [1.20416474, -0.9998551,0], [2.87830002, -0.99819343,0]]]
        # the second triangle is a line --> so area is 0, but 1 is the same but 3D --> in that sense it has an area
        triangles3D = [[[2, 1, 0], [1, 1, 1], [3, 1, 2]], [[2, 1, 0], [1, 1, 0], [3, 1, 0]],
                       [[2, 1, 0], [1, 1, 0], [3, 2, 0]], [[2, 1, 0], [1, 1, 0], [2, 2, 0]],
                       [[0, 0, 0], [3, 0, 0], [3, 3, 0]]]  # why 0 for the second triangle --> do I have a bug ???

        triangles2D = [[[2, 1], [1, 1], [3, 1]], [[2, 1], [1, 1], [3, 1]],
                       [[2, 1], [1, 1], [3, 2]], [[2, 1], [1, 1], [2, 2]],
                       [[0, 0], [3, 0], [3, 3]]]  # why 0 for the second triangle -

        heightmap_flat = np.asarray([[0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, 0],
                                     [0, 0, 0, 0, 0]])

        # print(heightmap_flat)

        heightmap_graded = np.asarray([[1, 2, 3, 4, 5],
                                       [6, 7, 8, 9, 10],
                                       [11, 12, 14, 15, 16],
                                       [17, 18, 19, 20, 21]])

        # print(heightmap_graded)

        # triangles2D= [[[2.37749672, -0.99445809], [1.20416474, -0.9998551], [2.87830002, -0.99819343]],[[2.37749672, -0.99445809], [1.20416474, -0.9998551], [2.87830002, -0.99819343]]]
        # if heightmap is provided --> get the coords from heightmap otherwise assume what is passed is already 3D
        # check and compare 2D and 3D
        # check also no dupes in the triangles!!!
        print(compute_3D_surfacearea(triangles3D, heightmap=None))  # heightmap_graded # None # heightmap_flat)z
        print(compute_3D_surfacearea(triangles2D, heightmap=None))
        print(compute_3D_surfacearea(triangles2D, heightmap=None,stop_on_2D=True))

        # if height map is provided then take it



# ça marche enfin et c'est pas trop dur à faire --> je peux améliorer ça je pense!!!
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys
from time import sleep

from qtpy.QtCore import QObject, QThread, Signal, Qt
# Snip...

# Step 1: Create a worker class
from qtpy.QtWidgets import QMainWindow, QApplication, QWidget, QLabel, QPushButton, QVBoxLayout
import traceback

from epyseg.tools.early_stopper_class import early_stop
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

class FakeWorker2(QObject):
    # loop = Signal(object)
    # progress = Signal(int)
    # finished = Signal()

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)

    def __init__(self, fn, *args, **kwargs):
        # QThread.__init__(self)
        early_stop.stop = False
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

        # self.signals = WorkerSignals()

        # Add the callback to our kwargs
        self.kwargs['progress_callback'] = self.progress


    def run(self):


        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.result.emit(result)  # Return the result of the processing
        finally:
            self.finished.emit()  # Done


    def stop(self):
        # self.terminate()
        # self.finished.emit()
        # self.signals.finished.emit()
        early_stop.stop = True
        logger.info('Stopping the process... (this may take time)')


class Worker2(QThread):
    # loop = Signal(object)
    # progress = Signal(int)
    # finished = Signal()

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)

    def __init__(self, fn, *args, **kwargs):
        QThread.__init__(self)
        early_stop.stop = False
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        # self.signals = WorkerSignals()

        # Add the callback to our kwargs
        self.kwargs['progress_callback'] = self.progress


    def run(self):


        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            try:
                traceback.print_exc()
                exctype, value = sys.exc_info()[:2]
                self.error.emit((exctype, value, traceback.format_exc()))
            except:
                pass
        else:
            self.result.emit(result)  # Return the result of the processing
        finally:
            self.finished.emit()  # Done


    def stop(self):

        # global stop_threads
        # stop_threads = True
        # def stop(self):

        # self.quit()
        # self.threadactive = False





        # self.terminate()
        # self.wait()

        early_stop.stop = True # let the code finish on its own
        logger.info('Stopping the process... (this may take time)')
        # EZDeepLearning.

        try:
            self.terminate()
        except:
            pass

        # qtimer = QTimer()
        # qtimer.setSingleShot(True)
        # # self.path.textChanged.connect(lambda: qtimer.start(600))
        # self.path.textChanged.connect(
        #     lambda x: qtimer.start(600))  # somehow my weird bug is here and only in some conditions
        # # # Is there a better way TODO that ????
        # # timer.timeout.connect(self._text_changed)
        # qtimer.timeout.connect(lambda: _thread.ter)
        # timer.singleShot(1000, this, SLOT(hardTerminate()))
        #
        # // Normal termination
        # _foo->terminate();
        #
        # // Waiting thread
        # _thread->wait();
        #
        # // Thread finished, cancel timer
        # timer.stop();
        # }
        # void hardTerminate()
        # {
        # // Ooops, normal termination failed, will terminate thread
        # _thread->terminate();
        # }


        # self.finished.emit()
        # self.signals.finished.emit()

    #
# class Worker(QObject):
#     finished = Signal()
#     progress = Signal(int)
#
#     def run(self):
#         """Long-running task."""
#         for i in range(10):
#             sleep(1)
#             self.progress.emit(i + 1)
#         self.finished.emit()
#
#     def stop(self):
#         self.terminate()

class Window(QMainWindow):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.thread = None
        self.setupUi()

    def setupUi(self):
        self.setWindowTitle("QThreadPool + QRunnable")
        self.resize(250, 150)
        self.centralWidget = QWidget()
        self.setCentralWidget(self.centralWidget)
        # Create and connect widgets
        self.stepLabel = QLabel("Hello, World!")
        self.stepLabel.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.longRunningBtn = QPushButton("Click me!")
        self.longRunningBtn.clicked.connect(self.runLongTask)

        self.stop = QPushButton("Stop me!")
        self.stop.clicked.connect(self.early_stop)

        # Set the layout
        layout = QVBoxLayout()
        layout.addWidget(self.stepLabel)
        layout.addWidget(self.longRunningBtn)
        layout.addWidget(self.stop)
        self.centralWidget.setLayout(layout)
    # Snip...
    def runLongTask(self):
        # Step 2: Create a QThread object
        # self.thread = Worker()
        # self.thread.setTerminationEnabled(True)
        # Step 3: Create a worker object
        self.thread = Worker2(self.loop)
        self.thread.setTerminationEnabled(True)
        # Step 4: Move worker to the thread
        # self.worker.moveToThread(self.thread)
        # Step 5: Connect signals and slots
        # self.thread.started.connect(self.worker.run)
        self.thread.finished.connect(self.thread.quit)
        # self.thread.finished.connect(self.thread.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.progress.connect(self.reportProgress)
        # Step 6: Start the thread
        self.thread.start()

        # Final resets
        self.longRunningBtn.setEnabled(False)
        self.thread.finished.connect(
            lambda: self.longRunningBtn.setEnabled(True)
        )
        self.thread.finished.connect(
            lambda: self.stepLabel.setText("Long-Running Step: 0")
        )


    def early_stop(self):
        if self.thread is not None:
            print('stopping')
            # self.thread.stop()
            # self.thread.quit()
            # self.thread.terminate()
            # self.thread.wait()
            self.thread.stop()

        self.thread = None

    def reportProgress(self, value):
        # TODO --> do that better
        print('progress', value)

    def loop(self, progress_callback):
        for i in range(100):
            # self.x = i
            # self.loop.emit(i)
            # self.progress.emit(i)
            if progress_callback is not None:
                progress_callback.emit(i)
            sleep(0.05)
        # self.finished.emit()
if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec())




# see how to best handle complex shapes and also allow a free mode with alignments and packing or stacking in x and y
# see how to edit shapes
# can a tree make the difference ???

# tt est parfait --> faut connecter le save et on est parti
# connect the save and exports
# sinon mettre dans un menu
# pas mal --> can allow zoom and alike
# TODO add also more menus around it --> TODO in order to handle better the stuff
# very good but why not showing anything ???
# ça marche pas mais je sais pas pkoi
# channels seem be fixed --> cool
# finalize whsed and corrections --> make sure they are channel dependent
# check which channels are being restored by the new wshed within the paint arena
# à faire
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import QScrollArea, QVBoxLayout, QWidget, QToolBar, QStatusBar, QHBoxLayout, QAction
import qtawesome as qta
from epyseg.draw.shapes.image2d import Image2D
from epyseg.figure.ezfig import MyWidget
from epyseg.figure.row import Row


# TODO --> could maybe add icons just below scroll if needed to add actions
# j'arrive pas à creer un truc dejà self embedded
# make it also handle other stuff

# --> nb replace the Createpaintwidget in multi_image_display.py by this --> better in fact
# shall I do full screen ???
# TODO handle channels too if needed
# could store mask and memory of drawing here too --> maybe on mouse release ???
# avec des ctrl+Z et des ctrl+R

# need zooms

class scrollable_EZFIG(QWidget):

    def __init__(self, custom_paint_panel=None):
        super().__init__()
        if custom_paint_panel is None:
            self.EZFIG_panel = MyWidget()
        else:
            self.EZFIG_panel = custom_paint_panel
        layout = QVBoxLayout()
        self.is_width_for_alternating = True
        # same as the other with embed in scroll area
        self.scrollArea = QScrollArea()
        self.scrollArea.setBackgroundRole(QPalette.Dark)
        self.scrollArea.setWidget(self.EZFIG_panel)
        self.scrollArea.setWidgetResizable(
            False)  # perfect --> do not change --> that is what I want in fact         # maybe not what I want finally maybe the set geometry is better in fact

        # see how the size is adjusted in paint and do the same here

        # MEGA TODO handle size for this object
        # self.EZFIG_panel.setFixedSize(QSize(300,300)) # in fact it was not showing because the widget object has no size --> see how to make it work better soon

        # do I need that and is this stuff having this object
        # self.EZFIG_panel.scrollArea = self.scrollArea
        layout.addWidget(self.scrollArea)
        self.setLayout(layout)
        status_bar = QStatusBar()
        layout.addWidget(status_bar)
        # self.EZFIG_panel.statusBar = status_bar
        # if False:
        self.add_shortcuts()
        self.drawing_commands()

    # this guy should handle the shortcuts it can and pass the rest to the other layer
    def add_shortcuts(self):
        # if True:
        #     return
        zoomPlus = QtWidgets.QShortcut("Ctrl+Shift+=", self)
        zoomPlus.activated.connect(self.zoomIn)
        zoomPlus.setContext(QtCore.Qt.ApplicationShortcut)

        # that works and disconnects all
        # zoomPlus.disconnect()

        zoomPlus2 = QtWidgets.QShortcut("Ctrl++", self)
        zoomPlus2.activated.connect(self.zoomIn)
        zoomPlus2.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus = QtWidgets.QShortcut("Ctrl+Shift+-", self)
        zoomMinus.activated.connect(self.zoomOut)
        zoomMinus.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus2 = QtWidgets.QShortcut("Ctrl+-", self)
        zoomMinus2.activated.connect(self.zoomOut)
        zoomMinus2.setContext(QtCore.Qt.ApplicationShortcut)

        ctrl0 = QtWidgets.QShortcut("Ctrl+0", self)
        ctrl0.activated.connect(self.zoom_reset)
        ctrl0.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrlS = QtWidgets.QShortcut("Ctrl+S", self)
        # self.ctrlS.activated.connect(self.EZFIG_panel.save)
        self.ctrlS.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrlM = QtWidgets.QShortcut("Ctrl+M", self)
        # self.ctrlM.activated.connect(self.EZFIG_panel.ctrl_m_apply)
        self.ctrlM.setContext(QtCore.Qt.ApplicationShortcut)

        self.shrtM = QtWidgets.QShortcut("M", self)
        # self.shrtM.activated.connect(self.EZFIG_panel.m_apply)
        self.shrtM.setContext(QtCore.Qt.ApplicationShortcut)

        # I can connect to progeny --> makes sense and really easy --> cool
        self.enterShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Return), self)
        # self.enterShortcut.activated.connect(self.EZFIG_panel.apply)
        self.enterShortcut.setContext(QtCore.Qt.ApplicationShortcut)

        self.enterShortcut2 = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Enter), self)
        # self.enterShortcut2.activated.connect(self.EZFIG_panel.apply)
        self.enterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        # en fait je peux aussi connecter des trucs au descendant --> pas bete et tres facile à faire!!!

        # pkoi ça ne marche pas !!!
        self.shiftEnterShortcut = QtWidgets.QShortcut("Shift+Enter", self)
        # self.shiftEnterShortcut.activated.connect(self.EZFIG_panel.shift_apply)
        self.shiftEnterShortcut.setContext(QtCore.Qt.ApplicationShortcut)
        #
        self.shiftEnterShortcut2 = QtWidgets.QShortcut("Shift+Return", self)
        # self.shiftEnterShortcut2.activated.connect(self.EZFIG_panel.shift_apply)
        self.shiftEnterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrl_shift_S_grab_screen_shot = QtWidgets.QShortcut('Ctrl+Shift+S', self)
        # self.ctrl_shift_S_grab_screen_shot.activated.connect(self.EZFIG_panel.grab_screen_shot)
        self.ctrl_shift_S_grab_screen_shot.setContext(QtCore.Qt.ApplicationShortcut)

        # peut etre faire un fullscreen shortcut --> can be useful

    def enable_shortcuts(self):
        self.disable_shortcuts()
        # self.ctrlS.activated.connect(self.EZFIG_panel.save)
        # self.ctrlM.activated.connect(self.EZFIG_panel.ctrl_m_apply)
        # self.shrtM.activated.connect(self.EZFIG_panel.m_apply)
        # self.enterShortcut.activated.connect(self.EZFIG_panel.apply)
        # self.enterShortcut2.activated.connect(self.EZFIG_panel.apply)
        # self.shiftEnterShortcut.activated.connect(self.EZFIG_panel.shift_apply)
        # self.shiftEnterShortcut2.activated.connect(self.EZFIG_panel.shift_apply)

    def disable_shortcuts(self):
        # TODO
        try:
            self.ctrlS.disconnect()
        except:
            pass
        # TODO
        try:
            self.ctrlM.disconnect()
        except:
            pass
        try:
            self.shrtM.disconnect()
        except:
            pass
        try:
            self.enterShortcut.disconnect()
        except:
            pass
        try:
            self.enterShortcut2.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut2.disconnect()
        except:
            pass

    # def enableMouseTracking(self):
    #     self.EZFIG_panel.drawing_enabled = True
    #
    # def disableMouseTracking(self):
    #     self.EZFIG_panel.drawing_enabled = False

    # ça marche --> ajouter fit to width ou fit to height
    def fit_to_width_or_height(self):
        if self.is_width_for_alternating:
            self.fit_to_width()
        else:
            self.fit_to_height()
        self.is_width_for_alternating = not self.is_width_for_alternating

    # almost there but small bugs
    def fit_to_width(self):
        # print('TODO implement that')
        # return
        # if self.EZFIG_panel.image is None:
        #     return
        # self.scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2
        width -= self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()
        width_im = self.EZFIG_panel.updateBounds().width()
        scale = width / width_im
        self.scaleImage(scale)

    def fit_to_height(self, nb=0):
        # print('TODO implement that')
        # return
        # paint = self.EZFIG_panels[nb]
        # if self.EZFIG_panel.image is None:
        #     return
        # scrollArea = self.scroll_areas[nb]
        height = self.scrollArea.height() - 2
        height -= self.scrollArea.horizontalScrollBar().sizeHint().height()
        height_im = self.EZFIG_panel.updateBounds().height()
        scale = height / height_im
        self.scaleImage(scale)

    def fit_to_window(self):
        # print('TODO implement that')
        # return
        # compute best fit
        # just get the size of the stuff and best fit it
        # paint = self.EZFIG_panels[nb]
        # if self.EZFIG_panel.image is None:
        #     return

        # QScrollArea.ensureWidgetVisible(
        # QScrollArea.ensureVisible
        # QScrollArea.setWidgetResizable
        # scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2  # required to make sure bars not visible
        height = self.scrollArea.height() - 2

        # scale image that it fits in --> change scale

        # width-=self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()

        height_im = self.EZFIG_panel.updateBounds().height()
        width_im = self.EZFIG_panel.updateBounds().width()
        scale = height / height_im
        if width / width_im < scale:
            scale = width / width_im
        self.scaleImage(scale)

    def zoomIn(self):
        # print('TODO implement that')
        # return
        # 10% more each time
        self.scaleImage(
            self.EZFIG_panel.scale + (self.EZFIG_panel.scale * 10.0 / 100.0))  # une hausse de 10% à chaque fois

    def zoomOut(self):
        # print('TODO implement that')
        # return
        # 10% less each time
        self.scaleImage(self.EZFIG_panel.scale - (self.EZFIG_panel.scale * 10.0 / 100.0))

    def zoom_reset(self):
        # print('TODO implement that')
        # return
        # resets zoom
        self.scaleImage(1.0)

    # pb this is redundant with the main GUI and not in sync --> can cause a lot of trouble --> dactivate for now
    def full_screen(self):
        if not self.isFullScreen():
            self.fullscreen.setIcon(qta.icon('mdi.fullscreen-exit'))
            self.fullscreen.setToolTip('Exit full screen')
            self.showFullScreen()
            # self.setWindowFlags(
            #     QtCore.Qt.Window |
            #     QtCore.Qt.CustomizeWindowHint |
            #     # QtCore.Qt.WindowTitleHint |
            #     # QtCore.Qt.WindowCloseButtonHint |
            #     QtCore.Qt.WindowStaysOnTopHint
            # )
        else:
            # settings = QSettings()
            # self.Stack.restoreGeometry(settings.value("geometry")) #.toByteArray()
            self.fullscreen.setIcon(qta.icon('mdi.fullscreen'))
            self.fullscreen.setToolTip('Enter full screen')
            self.setWindowFlags(QtCore.Qt.Widget)
            # # self.Stack.setWindowFlags(self.flags)
            # self.grid.addWidget(self.Stack, 0,
            #                     0)  # pas trop mal mais j'arrive pas à le remettre dans le truc principal
            # # dirty hack to make it repaint properly --> obviously not all lines below are required but some are --> need test, the last line is key though
            # self.grid.update()
            # self.Stack.update()
            # self.Stack.show()
            # self.centralWidget().setLayout(self.grid)
            # self.centralWidget().update()
            # self.update()
            self.show()
            self.repaint()
            # self.Stack.update()
            # self.Stack.repaint()
            # self.centralWidget().repaint()

    def scaleImage(self, scale):
        # print('TODO implement that')
        # return
        self.EZFIG_panel.set_scale(scale)
        # if self.EZFIG_panel.image is not None:
        # c'est ici la taille en fait
        #
        # self.EZFIG_panel.resize(self.EZFIG_panel.scale * self.EZFIG_panel.image.size())
        size = self.EZFIG_panel.updateBounds()
        size = QSize(size.width(), size.height())
        self.EZFIG_panel.resize(
            self.EZFIG_panel.scale * size)  # ça marche set ça qui gère la taille --> c'est ce que je veux
        # else:
        #     self.EZFIG_panel.resize(QSize(0, 0))
        self.EZFIG_panel.update()

    # can maybe add pen stuff here --> simpler and easier
    # need zoom buttons here too
    '''
    buttons/icons to add at some point
        "Cell smaller than this value will be destroyed"
        "small_cell_remover"
        
        "/Icons/refresh.png"
        "run 2 seeds watershed (click twice in a poorly segmented cell to have a the bond be created using the watershed algorithm)"
        
        "/Icons/Knob Snapback.png"
        "Undo"
        
        "/Icons/Knob Valid Blue.png"
        "remove small cells and apply correction"
        
        "/Icons/Knob Valid Green.png"
        "apply correction"
        
        "/Icons/channels.png"
        "Channel Selection"
        "Channel to use for Watershed, polarity, recentering,...")
        
        "/Icons/onebit_11.png"
        "Save"
        
        "/Icons/eye.png"
        "View below the mask (lasts 2s)"
        
        "/Icons/zoom in.png"
        "Zoom +"
        
        "/Icons/zoom out.png"
        "Zoom -"
        
        "/Icons/1in1.png"
        "1:1 aspect ratio"
        
        "/Icons/fit_2_screen.gif"
        "Fit to panel"
        
        "/Icons/double_diagonal_arrow.gif"
        "Alternate between best fit in height or in width"
        
        "/Icons/Import Picture Document.png"
        "Send Current View To The System Clipboard"
        
        "/Icons/pipet_small.png"
        "Color pipet"
        
        "Mask col"
        "mask color"
        
        "Correct Drift" 
    
    '''

    # add shortcut for show hide mask and maybe also full screen --> not a good idea

    def drawing_commands(self):
        # self.penSize = QSpinBox(objectName='penSize')
        # self.penSize.setToolTip('Change this value to change pen size')
        # self.penSize.setSingleStep(1)
        # self.penSize.setRange(1, 256)
        # self.penSize.setValue(3)
        # self.penSize.valueChanged.connect(self.penSizechange)

        # self.channels = QComboBox(objectName='channels')
        # self.channels.setToolTip('Select the channel of interest')
        # self.channels.addItem("merge")
        # self.channels.addItem("0")
        # self.channels.addItems(["1", "2", "3"])
        # self.channels.currentIndexChanged.connect(self.channelChange)

        hlayout_of_toolbars = QHBoxLayout()
        hlayout_of_toolbars.setAlignment(Qt.AlignLeft)
        self.tb1 = QToolBar()
        self.tb2 = QToolBar()
        # self.tb3 = QToolBar()

        # pencil_action = QAction(qta.icon('ei.pencil'), 'Pen Size', self)
        # pencil_action.setEnabled(False)
        # self.tb2.addAction(pencil_action)
        # # tb.addAction("sq...")
        # self.tb2.addWidget(self.penSize)

        # toolButton = QToolButton()
        # toolButton.setText("Draw")
        # toolButton.setIcon(self.style().standardIcon(QStyle.SP_DriveFDIcon))
        # tb.addWidget(toolButton)

        # fa5.save
        save_action = QAction(qta.icon('fa5.save'), 'Save mask', self)
        save_action.triggered.connect(self.EZFIG_panel.updateBounds)
        self.tb2.addAction(save_action)

        # toolButton2 = QToolButton()
        # toolButton2.setText("Draw")
        # toolButton2.setIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
        # tb.addWidget(toolButton2)

        # maybe reuse that to show skeleton un peu à la illustrator --> easy to find objects lost below others
        show_hide_mask = QAction(qta.icon('fa.bullseye'), 'Show/hide mask (same as pressing "M")', self)
        # show_hide_mask.triggered.connect(self.EZFIG_panel.m_apply)
        self.tb2.addAction(show_hide_mask)

        apply_drawing = QAction(qta.icon('fa.check'), 'Apply drawing (same as pressing the "Enter" key)',
                                          self)
        # apply_drawing.triggered.connect(self.EZFIG_panel.apply)
        self.tb2.addAction(apply_drawing)

        # toolButton3 = QToolButton()
        # toolButton3.setText("Draw")
        # toolButton3.setIcon(self.style().standardIcon(QStyle.SP_DialogCancelButton))
        # tb.addWidget(toolButton3)

        # toolButton4 = QToolButton()
        # toolButton4.setText("Draw")
        # toolButton4.setIcon(self.style().standardIcon(QStyle.SP_DialogCloseButton))
        # tb.addWidget(toolButton4)

        # toolButton5 = QToolButton()
        # toolButton5.setText("Draw")
        # toolButton5.setIcon(self.style().standardIcon(QStyle.SP_TrashIcon))
        # tb.addWidget(toolButton5)

        # toolButton6 = QToolButton()
        # toolButton6.setText("Draw")
        # toolButton6.setIcon(self.style().standardIcon(QStyle.SP_ArrowBack))
        # tb.addWidget(toolButton6)

        # toolButton7 = QToolButton()
        # toolButton7.setText("Draw")
        # toolButton7.setIcon(self.style().standardIcon(QStyle.SP_BrowserReload))
        # tb.addWidget(toolButton7)

        # self.small_cell_size = QSpinBox()
        # self.small_cell_size.setToolTip('Select the minimal pixel area for a cell to be considered as a cell.\nAny cell having an area less than this value will be removed.')
        # self.small_cell_size.setSingleStep(1)
        # self.small_cell_size.setRange(1, 1000000)
        # self.small_cell_size.setValue(10)
        # self.small_cell_size.valueChanged.connect(self.small_cell_size_changed)

        # self.tb2.addWidget(self.small_cell_size)
        # apply_rm_small_cells = QAction(qta.icon('mdi.check-underline'), 'Apply drawing and remove small cells (same as pressing "Shift+Enter")', self)
        # # apply_rm_small_cells.triggered.connect(self.EZFIG_panel.shift_apply)# self.EZFIG_panel.shift_apply
        # self.tb2.addAction(apply_rm_small_cells)

        # marche pas --> comment faire pr que ça marche...
        # local_reshed = QAction(qta.icon('mdi.reload'), 'Locally seeded watershed (same as pressing "Ctrl/Cmd + M")', self)
        # # local_reshed.triggered.connect(self.EZFIG_panel.ctrl_m_apply)# self.EZFIG_panel.shift_apply
        # self.tb2.addAction(local_reshed)
        # hgjhghj

        # toolButton8 = QToolButton()
        # toolButton8.setText("Zoom+")
        # toolButton8.setIcon(qta.icon('ei.zoom-in')) # mdi.magnify-plus-outline --> marche aussi
        # tb.addWidget(toolButton8)

        # toolButton9 = QToolButton()
        # toolButton9.setText("Zoom-")
        # toolButton9.setIcon(qta.icon('ei.zoom-out'))
        # tb.addWidget(toolButton9)

        zoom_plus = QAction(qta.icon('ei.zoom-in'), 'Zoom+', self)
        zoom_plus.triggered.connect(self.zoomIn)
        self.tb1.addAction(zoom_plus)

        zoom_minus = QAction(qta.icon('ei.zoom-out'), 'Zoom-', self)
        zoom_minus.triggered.connect(self.zoomOut)
        self.tb1.addAction(zoom_minus)

        zoom_width_or_height = QAction(qta.icon('mdi.fit-to-page-outline'),
                                                 'Alternates between best fit in width and height', self)
        zoom_width_or_height.triggered.connect(self.fit_to_width_or_height)
        self.tb1.addAction(zoom_width_or_height)

        zoom_0 = QAction(qta.icon('ei.resize-full'), 'Reset zoom', self)
        zoom_0.triggered.connect(self.zoom_reset)
        self.tb1.addAction(zoom_0)

        # redundant with TA full screen and not in sync --> disable for now
        self.fullscreen = QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        self.fullscreen.setToolTip('Enter full screen')
        self.fullscreen.triggered.connect(self.full_screen)
        self.tb1.addAction(self.fullscreen)

        # exit_fullscreen = QtWidgets.QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        # exit_fullscreen.triggered.connect(self.showFullScreen)
        # self.tb1.addAction(fullscreen)

        # self.channel_label = QLabel('Channels:')
        # self.channel_label.setToolTip('Select the channel of interest')
        # self.tb3.addWidget(self.channel_label)
        # self.tb3.addWidget(self.channels)

        # mdi.fullscreen
        # mdi.fullscreen-exit
        # ei.fullscreen --> a bit less good I find
        # fa.bullseye

        # could have show and hide mask here --> see how I can do that --> but do not delete it in fact

        # self.layout().addWidget(self.tb1)
        hlayout_of_toolbars.addWidget(self.tb2)

        # self.tb2.addAction('toto')
        # self.tb2.addAction('ttutu')
        # self.tb2.addAction('tata')
        # hlayout_of_toolbars.addWidget(self.tb3)

        # one for the zooms
        # one for drawing and apply different things
        # TODO do several toolbars to deactivate different levels of stuff
        # and only allow save in the appropriate stuff

        # self.tb3.addAction('toto')
        # self.tb3.addAction('ttutu')
        # self.tb3.addAction('tata')
        hlayout_of_toolbars.addWidget(self.tb1)

        self.layout().addLayout(hlayout_of_toolbars)

    def get_selected_channel(self):
        idx = self.channels.currentIndex()
        if idx == 0:
            return None
        else:
            return idx - 1

    # TODO also implement channel change directly within the display tool
    # if merge is applied --> apply on average of all channels --> maybe not so smart an idea but ok to start with and better than what I do in TA
    # def channelChange(self, i):
    #     # update displayed image depending on channel
    #     # dqqsdqsdqsd
    #     # pass
    #     # try change channel if
    #     self.EZFIG_panel.channelChange(i)
    #     # print('in channel change !!!')
    #     # if self.img is not None:
    #     #     # print('in', self.img.metadata)
    #     #     if self.Stack.currentIndex() == 0:
    #     #         # need copy the image --> implement that
    #     #         # print(self.img[..., i].copy())
    #     #         # print(self.img[..., i])
    #     #         if i == 0:
    #     #             self.EZFIG_panel.setImage(self.img)
    #     #             # print('original', self.img.metadata)
    #     #         else:
    #     #             # print('modified0', self.img.metadata)
    #     #             channel_img = self.img.imCopy(c=i - 1)  # it's here that it is affected
    #     #             # print('modified1', self.img.metadata)
    #     #             # print('modified2', channel_img.metadata)
    #     #             self.EZFIG_panel.setImage(channel_img)
    #     #         self.EZFIG_panel.update()
    #     # else:
    #     #     # logger.error("Not implemented yet TODO add support for channels in 3D viewer")
    #     #     # sdqdqsdsqdqsd
    #     #     self.loadVolume()

    # TODO maybe handle pen size change directly within the display --> can be useful by the way
    # def small_cell_size_changed(self):
    #     self.EZFIG_panel.minimal_cell_size = self.small_cell_size.value()
    #
    # # TODO maybe handle pen size change directly within the display --> can be useful by the way
    # def penSizechange(self):
    #     self.EZFIG_panel.brushSize = self.penSize.value()
    #     # self.update() # too slow if I update --> avoid
    #     # self.l1.setFont(QFont("Arial", size))
    #
    # def set_image(self, img):
    #
    #     # probably I need store raw image to avoid issues --> TODO ???
    #     self.EZFIG_panel.set_image(img)# bug seems to be here
    #
    #     img = self.EZFIG_panel.raw_image
    #     # print(type(img))
    #
    #     # else I need also reset channels
    #
    #     # if img is not None:
    #     #     if img.has_c():
    #     self._update_channels(img=img)
    #     # print('out')
    #     # channels = self.EZFIG_panel.get_nb_channels()
    #     # self._update_channels(channels)
    #     # make it update also the channels
    #
    # def set_mask(self, mask):
    #     self.EZFIG_panel.set_mask(mask)

    # def _update_channels(self, img):
    #     selection = self.channels.currentIndex()
    #     self.channels.disconnect()
    #     self.channels.clear()
    #     comboData = ['merge']
    #     if img is not None:
    #         try:
    #             if img.has_c():
    #                 for i in range(img.get_dimension('c')):
    #                     comboData.append(str(i))
    #         except:
    #             # assume image has no channel and return None
    #             # or assume last stuff is channel if image has more than 2
    #             pass
    #     # logger.debug('channels found ' + str(comboData))
    #     self.channels.addItems(comboData)
    #     # index = self.channels.findData(selection)
    #
    #     self.channels.currentIndexChanged.connect(self.channelChange)
    #     # print("data", index)
    #     if selection != -1 and selection < self.channels.count():
    #         self.channels.setCurrentIndex(selection)
    #         # self.channelChange(selection)
    #     else:
    #         self.channels.setCurrentIndex(0)
    #         # self.channelChange(0)

    # def freeze(self, bool, level=1):
    #
    #     self.tb1.setEnabled(True)
    #     self.tb2.setEnabled(True)
    #     self.tb3.setEnabled(True)
    #
    #     if level == 1:
    #         self.tb2.setEnabled(not bool)
    #     elif level == 2:
    #         self.tb2.setEnabled(not bool)
    #         self.tb3.setEnabled(not bool)
    #
    #     if bool:
    #         self.disable_shortcuts()
    #     else:
    #         self.enable_shortcuts()
    #     # remove draw mode maybe and freeze shortcuts


# pas mal faut juste voir ce que je peux faire
# peut etre permetre de mettre des custom panels

if __name__ == '__main__':
    # TODO add a main method so it can be called directly
    # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!
    import sys
    from qtpy.QtWidgets import QApplication

    # should probably have his own scroll bar embedded somewhere

    app = QApplication(sys.argv)

    # quite easy todo in fact

    # no clue why it always works just once

    # maybe make a scrollable TA paint that will handle and behave like that
    # maybe handle also stacks at some point ---> see how
    # --> almost all is ok now
    # class overriding_apply(Createpaintwidget):
    #     # all seems ok now and functions as in TA
    #     # just do the shift enter to get rid of small cells
    #
    #     # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
    #     def apply(self):
    #        self.apply_drawing(minimal_cell_size=0)
    #
    #     def shift_apply(self):
    #         # MEGA TODO IMPLEMENT SIZE within this stuff!!!
    #         self.apply_drawing(minimal_cell_size=10)
    #
    #     def ctrl_m_apply(self):
    #         self.manually_reseeded_wshed()
    #
    #     def save(self):
    #         self.save_mask()

    w = scrollable_EZFIG()  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
    # w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
    # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
    # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
    # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')
    # w.set_image('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1.tif')
    # w.set_mask('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1/handCorrection.tif')

    row1 = Row(Image2D('/E/Sample_images/sample_images_EZF/counter/00.png'),
               Image2D('/E/Sample_images/sample_images_EZF/counter/01.png'))
    row1.setX(0)
    row1.setY(0)

    row1.setToWidth(300)

    # print(row1.boundingRect())
    w.EZFIG_panel.shapes_to_draw = [row1]
    # w.EZFIG_panel.show()
    # w.EZFIG_panel.updateBounds() # --> it will update the size --> just need to do this in a smarter way but ok in fact
    # so indeed there is a size bug in the size of the row --> need a fix !!!
    w.EZFIG_panel.update_size()
    w.EZFIG_panel.update()
    # ça marche mais tt finalizer
    # that seems to work

    # w.freeze(True)
    # w.freeze(False)
    # w.freeze(True)

    # all is so easy to do this way

    # ça marche --> tres facile a tester
    # --> how can I do a zoom
    # ask whether self scroll or not --> if scrollable

    # mask = w.get_mask()
    #
    # plt.imshow(mask)
    # plt.show()
    print(w.EZFIG_panel.shapes_to_draw)  # --> ça marche en fait

    w.show()
    sys.exit(app.exec_())



import traceback
from scipy import ndimage
from skimage.filters import threshold_otsu
# from skimage.segmentation import watershed 
from skimage.segmentation import watershed
from epyseg.img import Img
from matplotlib import pyplot as plt
from skimage.measure import label, regionprops
from timeit import default_timer as timer
import os
import numpy as np
import statistics
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

def simpleFilter(img_orig, threshold=None, **kwargs): # threshold=None is autofilter
    if img_orig.has_c():
        for c in range(img_orig.shape[-1]):
            if threshold is not None:
                current = img_orig[...,c]
                current = current/current.max()
                current[current<=threshold]=0
                current[current>threshold]=255
                img_orig[...,c]=current
            else:
                current = img_orig[..., c]
                cur_threshold = threshold_otsu(current)
                current[current <= cur_threshold] = 0
                current[current > cur_threshold] = 255
                img_orig[..., c] = current
        return img_orig.astype(np.uint8)
    else:
        if threshold is not None:
            img_orig[img_orig <= threshold] = 0
            img_orig[img_orig > threshold] = 255
            return img_orig.astype(np.uint8)
        else:
            cur_threshold = threshold_otsu(img_orig)
            img_orig[img_orig <= cur_threshold] = 0
            img_orig[img_orig > cur_threshold] = 255
            return img_orig.astype(np.uint8)

def FilterMask(img_orig, final_wshed, filter='local median',  correction_factor=2,
             _DEBUG=False,
             _VISUAL_DEBUG=False,
             **kwargs):

    labs = label(Img.invert(final_wshed.astype(np.uint8)), connectivity=1, background=0)

    start = timer()

    output_folder = '/home/aigouy/Bureau/trash/test_new_seeds_seg_stuff/'
    if filter is None or filter == 0:
        return final_wshed.astype(np.uint8)
    else:
        if isinstance(filter, int):
            filter_by_size = filter
        else:
            filter_by_size = None
        avg_area = 0
        count = 0
        if _DEBUG:
            Img(final_wshed, dimensions='hw').save(os.path.join(output_folder, 'extras', 'test_size_cells.tif'))

        final_seeds = Img.invert(final_wshed)
        final_seeds = label(final_seeds, connectivity=1, background=0)

        if _VISUAL_DEBUG:
            plt.imshow(final_seeds)
            plt.show()

        removed_seeds = []
        keep_seeds = []

        labels_n_bbox = {}
        labels_n_area = {}
        border_cells = []
        ids_n_local_median = {}

        if isinstance(filter, str) and 'local' in filter:
            rps = regionprops(final_seeds)

            for region in rps:
                labels_n_bbox[region.label] = region.bbox
                labels_n_area[region.label] = region.area
                if (region.bbox[0] <= 3 or region.bbox[1] <= 3 or region.bbox[2] >= final_seeds.shape[-2] - 5 or
                        region.bbox[
                            3] >= \
                        final_seeds.shape[-1] - 5):
                    border_cells.append(region.label)

            _, tiles = Img.get_2D_tiles_with_overlap(final_seeds, overlap=64, dimension_h=-2, dimension_w=-1)

            for r in tiles:
                for tile in r:
                    rps2 = regionprops(tile)
                    for region in rps2:
                        if region.label in border_cells:
                            continue

                        if (region.bbox[0] <= 3 or region.bbox[1] <= 3 or region.bbox[2] >= final_seeds.shape[
                            -2] - 5 or
                                region.bbox[
                                    3] >= \
                                final_seeds.shape[-1] - 5):
                            continue

                        area_of_neighboring_cells = []
                        for region2 in rps2:
                            if region2.label == region.label:
                                continue
                            # find all cells with
                            if rect_distance(region.bbox, region2.bbox) <= 1:
                                area_of_neighboring_cells.append(labels_n_area[region2.label])

                        if area_of_neighboring_cells:
                            median = statistics.median_low(area_of_neighboring_cells)
                            ids_n_local_median[
                                region.label] = median / correction_factor
                            if region.area <= median / correction_factor:
                                removed_seeds.append(region.label)
                            else:
                                keep_seeds.append(region.label)
            removed_seeds = [x for x in removed_seeds if x not in keep_seeds]

            # TODO offer the things below as an option --> prevent removal of sure seeds or something like that
        else:
            areas = []

            for region in regionprops(final_seeds):
                if (region.bbox[0] <= 3 or region.bbox[1] <= 3 or region.bbox[2] >= final_seeds.shape[-2] - 5 or
                        region.bbox[3] >= final_seeds.shape[-1] - 5):
                    continue
                avg_area += region.area
                count += 1
                areas.append(region.area)
            avg_area /= count

            median = statistics.median_low(areas)

            if isinstance(filter, int):
                filter_by_size = filter
            elif 'avg' in filter:
                filter_by_size = avg_area / correction_factor
            elif 'median' in filter:
                filter_by_size = median / correction_factor
            # TODO maybe use stdev or alike to see if cell should really be removed
            if _DEBUG:
                print('filter cells below=', filter_by_size, 'avg cell area=', avg_area, 'median=',
                      median)  # , 'median', median

            if filter_by_size is not None and filter_by_size != 0:

                if _VISUAL_DEBUG:
                    plt.imshow(final_seeds)
                    plt.show()

                for region in regionprops(final_seeds):
                    labels_n_bbox[region.label] = region.bbox
                    labels_n_area[region.label] = region.area
                    if region.area < filter_by_size:
                        if (region.bbox[0] <= 2 or region.bbox[1] <= 2 or region.bbox[2] >= labs.shape[
                            -2] - 3 or
                                region.bbox[
                                    3] >= \
                                labs.shape[
                                    -1] - 3):
                            continue
                        removed_seeds.append(region.label)

        if _VISUAL_DEBUG:
            plt.imshow(final_seeds)
            plt.show()

        for region in regionprops(final_seeds):
            if region.label in removed_seeds:
                for coordinates in region.coords:
                    final_seeds[coordinates[0], coordinates[1]] = 0
        if _VISUAL_DEBUG:
            plt.imshow(final_seeds)
            plt.show()

        final_wshed = watershed(img_orig, markers=final_seeds, watershed_line=True)

        final_wshed[final_wshed != 0] = 1  # remove all seeds
        final_wshed[final_wshed == 0] = 255  # set wshed values to 255
        final_wshed[final_wshed == 1] = 0  # set all other cell content to
        if _VISUAL_DEBUG:
            plt.imshow(final_wshed)
            plt.show()

        duration = timer() - start
        if _DEBUG:
            print('final duration wshed in secs', duration)

        return final_wshed.astype(np.uint8)

def rect_distance(bbox1, bbox2):
    width1 = abs(bbox1[3] - bbox1[1])
    width2 = abs(bbox2[3] - bbox2[1])
    height1 = abs(bbox1[2] - bbox1[0])
    height2 = abs(bbox2[2] - bbox2[0])
    return max(abs((bbox1[1] + width1 / 2) - (bbox2[1] + width2 / 2)) - (width1 + width2) / 2,
               abs((bbox1[0] + height1 / 2) - (bbox2[0] + height2 / 2)) - (height1 + height2) / 2)


if __name__ == '__main__':
    pass



# a small GUI to load an ensemble/a series of models
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QRect, Qt, QRectF
from qtpy.QtWidgets import QWidget, QApplication, QGridLayout, QScrollArea, QVBoxLayout, QGroupBox, QPushButton
from epyseg.gui.open import OpenFileOrFolderWidget
import sys
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

class Ensemble_Models_Loader(QWidget):

    def __init__(self, parent_window=None):
        super().__init__(parent=parent_window)
        self.initUI()

    def initUI(self):
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.models_container = QWidget()

        self.scrollArea = QScrollArea()
        self.scrollArea.setWidgetResizable(True)

        # self.scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.models = []
        self.model_weights = []
        # add window to it --> TODO
        # self.input_model = OpenFileOrFolderWidget(parent_window=self, is_file=True,
        #                                           extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5 *.json *.JSON *.model);;All Files (*)",
        #                                           tip_text='Drag and drop a single model file here')
        # self.input_weights = OpenFileOrFolderWidget(parent_window=self, label_text='Load weights',
        #                                             is_file=True,
        #                                             extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5);;All Files (*)",
        #                                             tip_text='Drag and drop a single weight file here')  # TODO shall i add *.model ???
        #

        self.add_model_to_ensemble = QPushButton('+')
        self.add_model_to_ensemble.clicked.connect(self.add_new_model)

        self.remove_model_to_ensemble = QPushButton('-')
        self.remove_model_to_ensemble.clicked.connect(self.remove_last_model)


        self.models_container.layout = QVBoxLayout()
        # self.models_container.layout.setAlignment(Qt.AlignTop)
        # self.models_container.layout.setColumnStretch(0, 25)
        # self.models_container.layout.setColumnStretch(1, 75)
        # self.models_container.layout.setHorizontalSpacing(3)
        # self.models_container.layout.setVerticalSpacing(3)

        self.add_new_model()
        self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()
        # self.add_new_model()

        self.models_container.setLayout(self.models_container.layout)


        self.scrollArea.setWidget(self.models_container )

        # self.scrollArea.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        # self.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        # self.setFixedSize(self.size())

        layout.addWidget(self.scrollArea)
        layout.addWidget(self.add_model_to_ensemble)
        layout.addWidget(self.remove_model_to_ensemble)

        self.setLayout(layout)
        self.setFixedHeight(200)

    def add_new_model(self):
        self.groupBox_model = QGroupBox('Model '+str(len(self.models)), objectName='Model_'+str(len(self.models)))
        self.groupBox_model.setEnabled(True)
        # groupBox layout
        self.groupBox_model_layout = QVBoxLayout()
        # self.groupBox_model_layout.setAlignment(Qt.AlignTop)
        # self.groupBox_model_layout.setColumnStretch(0, 25)
        # self.groupBox_model_layout.setColumnStretch(1, 25)
        # self.groupBox_model_layout.setColumnStretch(2, 50)
        # self.groupBox_model_layout.setColumnStretch(3, 2)
        # self.groupBox_model_layout.setHorizontalSpacing(3)
        # self.groupBox_model_layout.setVerticalSpacing(3)

        # add a groupbox per model

        # TODO maybe allow this to be stored and or reupdated --> not so easy would need to design specific stuff
        #TODO should I allow it to be saved in ini, most likely not...
        self.input_model = OpenFileOrFolderWidget(parent_window=self,  label_text='Load model',is_file=True,
                                                  extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5 *.json *.JSON *.model);;All Files (*)",
                                                  tip_text='Drag and drop a single model file here')
        self.input_weights = OpenFileOrFolderWidget(parent_window=self, label_text='Load weights (otpional)',
                                                    is_file=True,
                                                    extensions="Supported Files (*.h5 *.H5 *.hdf5 *.HDF5);;All Files (*)",
                                                    tip_text='Drag and drop a single weight file here')  # TODO shall i add *.model ???


        self.models.append(self.input_model)
        self.model_weights.append(self.input_weights)

        # add it to the layout too

        self.groupBox_model_layout.addWidget(self.input_model)
        self.groupBox_model_layout.addWidget(self.input_weights)

        self.groupBox_model.setLayout(self.groupBox_model_layout)

        self.models_container.layout.addWidget(self.groupBox_model)


    def remove_last_model(self):
        # for i in range(self.models_container.layout.count()):
        #     print(i)
        # for i in reversed(range(models_container.count())):
        try:
            self.models_container.layout.itemAt( self.models_container.layout.count()-1).widget().setParent(None)
        except:
            pass
        # self.models_container.layout.itemAt(self.models_container.layout.count()-1).widget().setParent(None)

        if self.models:
            self.models.pop()
        if self.model_weights:
            self.model_weights.pop()

    def get_models(self):
        # loop of other model lists
        if not self.model_weights:
            logger.error('No model --> nothing can be done, sorry...')
            return None
        # else return path to models and weights
        models_and_weights = {}
        for iii,model in enumerate(self.models):
            if model.text() is not None:
                models_and_weights[iii]=[model.text(), self.model_weights[iii].text()]
        if not models_and_weights:
            logger.error('No model loaded --> nothing can be done, sorry...')
            return None
        return models_and_weights


if __name__ == '__main__':
    # just for a test
    app = QApplication(sys.argv)
    ex = Ensemble_Models_Loader()
    ex.show()
    app.exec_()

    # that seems to work
    print(ex.get_models())



# seems perfect --> gives no swapping --> just errors due to dupes and in that case I should just take the most likely cell --> cell with closest area or closest shape based on its vertices --> I'm almost done!!!!

# I think this shit is now fully coded...


# https://github.com/voxelmorph/voxelmorph --> is that what I was looking for ???
# maybe what I was looking for https://colab.research.google.com/drive/1WiqyF7dCdnNBIANEY80Pxw_mVz4fyV-S?usp=sharing --> bingo I found you THIS IS JUST EXACTLY WHAT I WANTED / NEEDED AND IT'S TENSORFLOW 2.0!!!!!!!!!!


# nb the IJ sift align works very well too but couldn't find an easy python equivalent...


# apparement qq un y a deja pensé https://pubmed.ncbi.nlm.nih.gov/18267377/
# http://bigwww.epfl.ch/publications/thevenaz9801.html


# the idea is to crop further and further the image in order to get pyramidal correction of seefd position in order to get the correct tracking done

# get an image chop it into tiles then chop each registered tile further and register it on the big image or change the stuff


# or do smaller things
# or detect a swap and if a swap is detected then try recompute the translation

# or chop new image into small tiles and try to match each of these tiles on the bigger one

# try combinations of these things

# think about it


# try register a small region on the big --> how would that work if images don't have the same size

# ask for depth of tiling


# from skimage.feature import register_translation
from skimage.registration import phase_cross_correlation # change in scipy 0.19
# image = data.camera()
from skimage import transform
import numpy as np

# really very good

# parameters

# threshold
# threshold_translation = 64  # all the further translations should be small --> if that is not the case then there has to be a pb

__DEBUG__ = False




def apply_translation(img, y,x):
    afine_tf = transform.AffineTransform(translation=(x, y))
    translated = transform.warp(img, inverse_map=afine_tf, order=0, preserve_range=True)
    return translated

def pre_register_images(orig_t0, orig_t1, apply_to_orig_t0=False):
    global_translation, error, diffphase = phase_cross_correlation(orig_t0, orig_t1)
    gloabl_t0 = -global_translation[0]
    gloabl_t1 = -global_translation[1]
    if not apply_to_orig_t0:
        return gloabl_t0, gloabl_t1
    else:
        return apply_translation(orig_t0, -gloabl_t0, -gloabl_t1)

def get_pyramidal_registration(orig_t0, orig_t1, depth=2, threshold_translation=20):
    # orig_t0 = Img('/E/Sample_images/tracking_test/leg_cropped/100807_leg4_male.lei - leg4_male_Series031.tif')#[455:455+122,544:544+127] #[384:512,512:640]
    # orig_t1 = Img('/E/Sample_images/tracking_test/leg_cropped/100807_leg4_male.lei - leg4_male_Series033.tif')#[455:455+122,544:544+127] #[384:512,512:640]

    translation_matrix = np.zeros((*orig_t0.shape, 2))

    # global_translation, error, diffphase = register_translation(orig_t0, orig_t1)
    #
    # if __DEBUG__:
    #     print('error', error,
    #           diffphase)  # if error is bigger upon further registration --> try ignore it or use a ROI to register only the region of interest --> similar to what I had in TA...
    #
    # gloabl_t0 = -global_translation[0]
    # gloabl_t1 = -global_translation[1]

    gloabl_t0, gloabl_t1 = pre_register_images(orig_t0, orig_t1)

    translation_matrix[..., 0] += gloabl_t0
    translation_matrix[..., 1] += gloabl_t1

    # small bug :
    '''
    /home/aigouy/.local/lib/python3.7/site-packages/skimage/feature/register_translation.py:104: RuntimeWarning: invalid value encountered in true_divide
      (src_amp * target_amp)
    '''

    height = orig_t0.shape[0]
    width = orig_t0.shape[1]

    # Nb i need get parent translation and further apply it before retracking the stuff --> need get the cropping

    block_size = 256

    for i in range(depth):
        try:
            # final_value = i + (pyramidal_skip if i > 0 else 0)
            tile_height = block_size // (i * 2)
            tile_width = block_size // (i * 2)
        except:
            tile_height = block_size
            tile_width = block_size

        if __DEBUG__:
            print('size of stuff', ((i + 1) * 2), tile_width)
        # trans_y = []
        for y in range(0, height, tile_height):
            # trans_x = []
            for x in range(0, width, tile_width):
                if __DEBUG__:
                    print(y, x, '-->', y + tile_height, x + tile_width)
                # take bigger for one and not for the other --> include surrounding inj one and black in the other
                # tile_t0=mask_t0[y:y+tile_height,x:x+tile_width]
                # tile_t1=mask_t1[y:y+tile_height,x:x+tile_width]
                # translate image 0 before further registering it
                afine_tf = transform.AffineTransform(
                    translation=(-translation_matrix[y, x, 1], -translation_matrix[y, x, 0]))
                translated = transform.warp(orig_t0, inverse_map=afine_tf, order=0, preserve_range=True)

                # plt.imshow(translated)
                # plt.show()

                tile_t0 = translated[y:y + tile_height, x:x + tile_width]
                tile_t1 = orig_t1[y:y + tile_height, x:x + tile_width]
                # t0, t1 = translation(tile_t0, tile_t1)

                # can I create a patchwork of the registered image...


                # shift, error, diffphase = register_translation(tile_t0, tile_t1)
                # change for v 0.19 of scipy
                shift, error, diffphase = phase_cross_correlation(tile_t0, tile_t1)
                t0 = -shift[0]
                t1 = -shift[1]
                # imshow(mask_t0, mask_t1, im2)

                # print('translation', t0, t1, scale, angle)  # seems to be what I want --> just try default TA algo on that --> TODO
                if __DEBUG__:
                    print('translation', t0, t1)

                # ignore too big translations because they are most likely artifacts
                if abs(t0) > threshold_translation or abs(t1) > threshold_translation:
                    if __DEBUG__:
                        print('most likely an erroneous translation detected --> ignoring', t0, t1)
                    continue

                translation_matrix[y:y + tile_height, x:x + tile_width, 0] += t0
                translation_matrix[y:y + tile_height, x:x + tile_width, 1] += t1
                # trans_x.append((t0, t1))
            # trans_y.append(trans_x)

        # print('trans_y', trans_y)
        # in a way it does compute a flowfield too...

    return translation_matrix



#https://docs.python.org/2/library/operator.html

import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from epyseg.draw.shapes.rect2d import Rect2D
from qtpy.QtCore import QPointF, QRectF
# logger
from epyseg.tools.logger import TA_logger
from epyseg.figure.row import Row  # KEEP Really required to avoid circular imports
logger = TA_logger()


# TODO hack it that it can handle rows or images if needed

class col(Rect2D):

    def __init__(self, *args, space=3, width=None):
        self.images = []
        self.widthInPixel = width
        self.space = space
        for arg in args:
        #     if isinstance(arg, Column):
        #         arg.packX(space=space)
        #         arg.setToWidth(self.widthInPixel)
            self.images.append(arg)
            # else:
            #     self.images.append(Column(arg, space=space, width=width))
        super(Rect2D, self).__init__()
        self.packY(space=space)

    def __or__(self, other):
        if isinstance(other, col):
            pos = self.get_P1()
            other_pos = other.get_P1()
            self.set_P1(other_pos)
            other.set_P1(pos)
        else:
            logger.error('swapping not implemented yet for ' + str(type(other)))

    def __sub__(self, other):
        if other in self.images:
            self.images.remove(other)
            return
        for row in self.images:
            row -= other
        for row in reversed(self.images):
            if row.isEmpty():
                self.images.remove(row)

        return self

    def __add__(self, other):
        from epyseg.figure.row import Row  # KEEP Really required to avoid circular imports # TODO is that still true ?
        if isinstance(other, Row):
            other.packX(space=self.space)
            other.setToWidth(self.widthInPixel)
            self.images.append(other)
            return self
        self.images.append(Row(other, space=self.space, width=self.widthInPixel))
        self.packY(space=self.space)
        return self

    def __len__(self):
        if self.images is None:
            return 0
        return len(self.images)

    def __iter__(self):
        ''' Returns the Iterator object '''
        self._index = -1
        return self

    def __next__(self):
        ''' Returns the next row of the col '''
        self._index += 1
        if self._index < len(self.images):
            return self.images[self._index]
        # End of Iteration
        raise StopIteration

    def setWidthInPixel(self, widthInPixel):
        self.widthInPixel = widthInPixel
        self.packY()

     # Forces the block to be of width (width_in_px)
    def setToWidth(self, width_in_px):
        for image in self:
            image.setToHeight(width_in_px)
        self.packY(space=self.space)

    def setToHeight(self, height_in_px):
        for image in self:
            image.setToHeight(height_in_px)
        self.packY(space=self.space)

    def getIncompressibleWidth(self):
        extra_space = 0;
        return extra_space;

    def packX(self, space=3):
        # pack shapes in x

        last_x = 0
        last_y = 0

        for i in range(len(self.images)):
            row = self.images[i]
            if i != 0:
                last_x += space
            row.set_P1(last_x, row.get_P1().y())
            # get all the bounding boxes and pack them with desired space in between
            # get first point and last point in x
            x = row.boundingRect().x()
            y = row.boundingRect().y()
            last_x = row.boundingRect().x() + row.boundingRect().width()
            last_y = row.boundingRect().y() + row.boundingRect().height()
        self.updateBoudingRect()


    def packY(self, space=3):

        if space is None:
            space = 0

        last_x = 0
        last_y = 0


        for i in range(len(self.images)):
            row = self.images[i]
            if i != 0:
                last_y += space
            row.set_P1(row.get_P1().x(), last_y)

            x = row.boundingRect().x()
            y = row.boundingRect().y()
            last_x = row.boundingRect().x() + row.boundingRect().width()
            last_y = row.boundingRect().y() + row.boundingRect().height()
        self.updateBoudingRect()

    def updateBoudingRect(self):
        '''updates the image bounding rect depending on content

        '''
        x = None
        y = None
        x2 = None
        y2 = None
        for row in self:
            topLeft = row.get_P1()
            if x is None:
                x = topLeft.x()
            if y is None:
                y = topLeft.y()
            x = min(topLeft.x(), x)
            y = min(topLeft.y(), y)
            if x2 is None:
                x2 = topLeft.x() + row.boundingRect().width()
            if y2 is None:
                y2 = topLeft.y() + row.boundingRect().height()
            x2 = max(topLeft.x() + row.boundingRect().width(), x2)
            y2 = max(topLeft.y() + row.boundingRect().height(), y2)

        if x is not None:
            self.setX(x)
        if y is not None:
            self.setY(y)
        if x2 is not None:
            self.setWidth(x2 - x)
        if y2 is not None:
            self.setHeight(y2 - y)

    def set_P1(self, *args):
        curP1 = self.get_P1()
        Rect2D.set_P1(self, *args)
        newP1 = self.get_P1()
        # need do a translation of all its content images
        for row in self:
            row.translate(newP1.x() - curP1.x(), newP1.y() - curP1.y())

    def translate(self, *args):
        if len(args)==1:
            point = args[0]
            QRectF.translate(self, point.x(), point.y())
            for img in self:
                img.translate(point.x(), point.y())
        else:
            QRectF.translate(self, args[0], args[1])
            for img in self:
                img.translate(QPointF(args[0], args[1]))


    def draw(self, painter, draw=True):
        if draw:
            painter.save()
        if draw:
            counter = 0
            for row in self:
                print('drawing', counter, row.boundingRect())
                counter+=1
                row.draw(painter, draw=draw)
            painter.restore()

    def fill(self, painter, draw=True):
        if self.fill_color is None:
            return
        if draw:
            painter.save()
        if draw:
            for row in self:
                row.fill(painter, draw=draw)
            painter.restore()

        # TODO pb will draw the shape twice.... ---> because painter drawpolygon is called twice

    def drawAndFill(self, painter):
        painter.save()
        self.draw(painter, draw=True)
        painter.restore()

if __name__ == '__main__':
    from epyseg.draw.shapes.image2d import Image2D  # KEEP Really required to avoid circular imports

    img1 = Image2D('./../data/counter/00.png')
    img2 = Image2D('./../data/counter/01.png')
    img3 = Image2D('./../data/counter/02.png')
    img4 = Image2D('./../data/counter/03.png')
    img5 = Image2D('./../data/counter/04.png')
    img6 = Image2D('./../data/counter/05.png')
    img8 = Image2D('./../data/counter/06.png')
    img9 = Image2D('./../data/counter/07.png')
    img10 = Image2D('./../data/counter/08.png')
    img11 = Image2D('./../data/counter/09.png')

    result = img1 + img2

    print(result)
    print(len(result.images))

    #    result += img3
    #    result += img4

    #    print(result)
    #    print(len(result.images)) # ça ça marche

    #    if True:
    #        sys.exit(0)

    # img3.pop()

    row2 = img3 + img4 + img8
    # row2 = row(img3, img4)

    print(row2)
    print(row2.images)  # why 4 images here ....
    print('row2', len(row2.images))  # pas bon trop d'images --> pkoi ????

    print(Image2D)

    fig = result + row2
    print(fig)
    print(len(fig.images))

    count = 0
    for img in fig.images:
        print(count, img.filename)
        count += 1

    final_result = fig - img4
    print(final_result)
    print(final_result.images)
    print(len(final_result.images))

    # ça marche

    final_result = img4 + fig  # ça ne marche pas
    print(final_result)
    print(final_result.images)
    print(len(final_result.images))

    # si je divide une row alors ça ajoute une colonne à la figure --> return 2 rows in a figure



    fig = img5 / img6

    print(fig)
    print(fig.rows)
    print(len(fig.rows))

    final_fig = fig - img5

    print(final_fig)
    print(final_fig.rows)
    print(len(final_fig.rows))

    final_fig += row2

    print(final_fig)
    print(final_fig.rows)
    print(len(final_fig.rows))

    fig = (img9) / (img10 + img11)
    print(fig)
    print(fig.rows)
    print(len(fig.rows))
    print('r1', len(fig.rows[0].images))
    print('r2', len(fig.rows[1].images))

    # try the pack and check it
    # almost there now in fact...

    for img in row2:
        print(img.boundingRect())

    print(row2)
    row2.packX()
    for img in row2:
        print(img.boundingRect())

    row2.packY()
    for img in row2:
        print(img.boundingRect())

    # ça marche

    # tester le draw sur une image --> balancer un qimage ???



    # print('#'*20)
    # from qtpy.QtCore import QRectF, QPointF
    # test = QRectF(0,0,411,512)
    # print(test)
    # test.setX(414) # --> fucks width --> need reput it in correct position
    # # test.setWidth(411)
    # print(test)
    # test = QRectF(0,0,411,512)
    # print(test)
    # test.setTopLeft(QPointF(414,0))
    # print(test)



# TODO handle selection and all the fusion options and also the normal save without export
# do a graph editor embedded --> see how --> make it as simple as possible and make the macro language so simple
# find an easy way to convert any action to a simple macro code --> would be great if I could save some simple code just for the fun of it
# think about it but I guess it will be easy
# store all the macro commands

# first thing to do is to get the ctrl clicks

# TODO implement the snap --> not so hard I guess if I drop on nothing

# icons to use at some point maybe:
# mdi.format-paint
# fa5s.paint-brush
# fa.file-picture-o
# fa.newspaper-o
# mdi.pdf-box
# mdi.export
# fa.folder-open-o
# ei.graph
# fa.paragraph
# fa5s.tools
# mdi.tools
# mdi.information-outline
# mdi.format-vertical-align-center
# mdi.format-vertical-align-top
# mdi.format-underline
# mdi.format-text-rotation-up

# pas mal --> good start for a GUI but add menus too


# this is the full EZF GUI that will contain everything needed for the plot
# need a menu a main panel and a side bar
# or many side bars --> think about it and keep it as simple as possible
# create a dynamic demo with an easy possibility to launch it so that the user can easily learn things/could have a demo tab
# see how to serialize everything --> maybe in the form of a script
# --> would be very easy to do I guess
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import copy
import traceback

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtGui import QColor, QTextCursor, QTextCharFormat
from qtpy.QtWidgets import QApplication, QStackedWidget, QWidget, QTabWidget, QScrollArea, QVBoxLayout, QPushButton, \
    QGridLayout, QTextBrowser, QFrame, QProgressBar, QGroupBox, QAction
import qtawesome as qta
import logging
import os

from epyseg.dialogs.opensave import saveFileDialog
from epyseg.draw.shapes.circle2d import Circle2D
from epyseg.draw.shapes.ellipse2d import Ellipse2D
from epyseg.draw.shapes.freehand2d import Freehand2D
from epyseg.draw.shapes.image2d import Image2D
from epyseg.draw.shapes.line2d import Line2D
from epyseg.draw.shapes.point2d import Point2D
from epyseg.draw.shapes.polygon2d import Polygon2D
from epyseg.draw.shapes.polyline2d import PolyLine2D
from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.draw.shapes.scalebar import ScaleBar
from epyseg.draw.shapes.square2d import Square2D
from epyseg.draw.shapes.txt2d import TAText2D
from epyseg.draw.shapes.vectorgraphics2d import VectorGraphics2D
from epyseg.figure.column import Column
from epyseg.figure.row import Row
from epyseg.figure.scrollableEZFIG import scrollable_EZFIG
from epyseg.tools.qthandler import XStream, QtHandler
from epyseg.uitools.blinker import Blinker
# from epyseg.ta.GUI.pyta import Overlay
from epyseg.ta.GUI.overlay_hints import Overlay

from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()


__MAJOR__ = 0
__MINOR__ = 1
__MICRO__ = 0
__RELEASE__ = 'b'  # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers --> alpha beta, ...
__VERSION__ = ''.join(
    [str(__MAJOR__), '.', str(__MINOR__), '.'.join([str(__MICRO__)]) if __MICRO__ != 0 else '', __RELEASE__])
__AUTHOR__ = 'Benoit Aigouy'
__NAME__ = 'PyFig: A Scientific Figure Creation Assistant' # TODO think about a name and how to handle that
__EMAIL__ = 'baigouy@gmail.com'


# DEBUG = False
DEBUG = True

class EZFIG_GUI(QtWidgets.QMainWindow):
    # stop_threads = False

    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

        # global early_stop
        # early_stop = False
        self.setAcceptDrops(True)  # we accept DND but just to transfer it to the current list handlers!!!

    def initUI(self):

        try:
            screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
            centerPoint = QApplication.desktop().screenGeometry(screen).center()
        except:
            from qtpy.QtGui import QGuiApplication
            centerPoint = QGuiApplication.primaryScreen().geometry().center()

        # should fit in 1024x768 (old computer screens)
        window_width = 900
        window_height = 700
        self.setGeometry(
            QtCore.QRect(centerPoint.x() - int(window_width / 2), centerPoint.y() - int(window_height / 2),
                         window_width,
                         window_height))

        # set the window icon
        # self.setWindowIcon(QtGui.QIcon('./../../IconsPA/src/main/resources/Icons/ico_packingAnalyzer2.gif'))
        # variantMap =QVariant()
        # variantMap.insert("color", QColor(10, 10, 10));



        self.setWindowIcon(qta.icon('fa.newspaper-o',
                                    color=QColor(200, 200, 200)))  # very easy to change the color of my icon

        # zoom parameters

        self.setWindowTitle(__NAME__ + ' v' + str(__VERSION__))

        # this is the default paint window of TA
        self.paint = scrollable_EZFIG()
        # this is the properties window (allows to set useful parameters that can be used by TA

        self.Stack = QStackedWidget(self)



        # Initialize tab screen
        self.tabs = QTabWidget(self)
        self.tab0 = QWidget()
        # self.tab0.setToolTip("jsdhqsjkdhkj sdqsh d")
        self.tab1 = QWidget()


        # Add tabs
        self.tabs.addTab(self.tab0, 'Pre-process')  # contains stuff to do max projs and heightmaps from images
        self.tabs.setTabToolTip(0,'Use a CARE model to perform surface extraction/create a 2D projection from a 3D epithelial stack.')
        self.tabs.addTab(self.tab1, 'Segmentation')
        self.tabs.setTabToolTip(1, 'Segment epithelial images.')
        # self.tabs.addTab(self.tab2, 'Analysis')

        self.tabs.currentChanged.connect(self.onTabChange)

        # force seg tab to appear first
        # self.tabs.setCurrentIndex(1)

        # Create first tab
        self.tab0.layout = QVBoxLayout()
        self.surf_proj = QPushButton("Surface projection")
        self.surf_proj.setToolTip(
            "Extracts, using a CARE model without the denoising module,\nthe apical fluorescent signal from a 3D image of an epithelium and creates a height map")
        # self.surf_proj.clicked.connect(self.surf_proj_run)
        self.tab0.layout.addWidget(self.surf_proj)
        # self.denoising_surf_proj = QPushButton("Denoising surface projection")
        # self.denoising_surf_proj.setToolTip(
        #     "Same as \"Surface projection\" using the complete CARE model, denoiser included.\nBe aware that the denoising module can generate unexpected/unwanted results,\nespecially, when the input data significantly differs from the data used to train the model.\nSo, altogether, denoising/image reconstruction is not necessarily improving the quality of the 2D projection.")
        # self.denoising_surf_proj.clicked.connect(self.surf_proj_run)
        # self.tab0.layout.addWidget(self.denoising_surf_proj)
        self.tab0.setLayout(self.tab0.layout)

        # Create first tab
        self.tab1.layout = QVBoxLayout()

        # self.pushButton0 = QPushButton("Deep Learning/EPySeg segmentation")
        # self.pushButton0.clicked.connect(self.epyseg_seg)
        # self.pushButton0.setToolTip("Segment an epithelium using the EPySeg model")
        # self.tab1.layout.addWidget(self.pushButton0)

        # self.pushButton1 = QPushButton("Watershed segmentation")
        # self.pushButton1.clicked.connect(self.run_watershed)
        # self.pushButton1.setToolTip("Segment an epithelium using the wahtershed algorithm")
        # self.tab1.layout.addWidget(self.pushButton1)
        self.tab1.setLayout(self.tab1.layout)
        # self.pushButton2 = QPushButton("Deep learning segmentation")
        # self.pushButton2.clicked.connect(self.deep_learning)
        # self.tab1.layout.addWidget(self.pushButton2)



        # content_widget.layout = QVBoxLayout()
        # sub_layout_preview = QGridLayout()
        # sub_layout_preview.setColumnStretch(0, 10)
        # sub_layout_preview.setColumnStretch(1, 90)
        #
        # preview_img_label = QLabel("Image/Data")
        # sub_layout_preview.addWidget(preview_img_label, 0, 0)
        #
        # # content_widget.layout.addWidget(preview_img_label)
        # self.image_preview_combo = QComboBox()
        # self.image_preview_combo.currentTextChanged.connect(self.preview_changed)
        # sub_layout_preview.addWidget(self.image_preview_combo, 0, 1)
        # content_widget.layout.addLayout(sub_layout_preview)
        #
        # self.groupBox_color_coding = QGroupBox('Color coding:')
        # self.groupBox_color_coding.setCheckable(True)
        # self.groupBox_color_coding.setChecked(False)
        # self.groupBox_color_coding.toggled.connect(self.preview_changed)
        #
        # color_coding_layout = QHBoxLayout()
        # color_coding_layout.setContentsMargins(0, 0, 0, 0)
        # color_coding_layout.setAlignment(Qt.AlignLeft)
        # # no local global
        # # color_coding_label = QLabel('Color coding: ')
        # # color_coding_layout.addWidget(color_coding_label, 0, 0)
        #
        # # self.radioButton1 = QRadioButton('No')
        # self.radioButton2 = QRadioButton('Current/Local')
        # self.radioButton2.setToolTip("Apply local (only for the current frame) color coding")
        # self.radioButton3 = QRadioButton('Global')
        # self.radioButton3.setToolTip("Apply global color coding (i.e. max and min is the max and min of all frames)")
        # # self.radioButton3.setEnabled(False)  # not supported yet --> reactivate it when it will be supported # see if I can show a scalebar somewhere in the stuff
        # self.radioButton2.setChecked(True)
        #
        # # self.radioButton1.toggled.connect(self.preview_changed)
        # self.radioButton2.toggled.connect(self.preview_changed)
        # self.radioButton3.toggled.connect(self.preview_changed)
        #
        # # color_coding_layout.addWidget(self.radioButton1, 0, 1)
        # color_coding_layout.addWidget(self.radioButton2)
        # color_coding_layout.addWidget(self.radioButton3)
        #
        # predict_output_radio_group = QButtonGroup()
        # # predict_output_radio_group.addButton(self.radioButton1)
        # predict_output_radio_group.addButton(self.radioButton2)
        # predict_output_radio_group.addButton(self.radioButton3)
        #
        # self.excluder_label = QCheckBox('Exclude % lower and upper outliers: ')
        # self.excluder_label.setEnabled(True)
        # self.excluder_label.setChecked(False)
        #
        # self.excluder_label.stateChanged.connect(self.upper_or_lower_limit_changed)
        #
        # # TODO enable or disable the stuffs along with it
        # color_coding_layout.addWidget(self.excluder_label)
        #
        # self.upper_percent_spin = QDoubleSpinBox()
        # self.upper_percent_spin.setEnabled(False)
        # self.upper_percent_spin.setRange(0., 0.49)
        # self.upper_percent_spin.setSingleStep(0.01)
        # self.upper_percent_spin.setValue(0.05)
        # # self.upper_percent_spin.valueChanged.connect(self.upper_or_lower_limit_changed)
        # self.upper_percent_spin.valueChanged.connect(lambda x: delayed_preview_update.start(600))
        #
        # self.lower_percent_spin = QDoubleSpinBox()
        # self.lower_percent_spin.setEnabled(False)  # TODO code and enable this soon but ok for now
        # self.lower_percent_spin.setRange(0., 0.49)
        # self.lower_percent_spin.setSingleStep(0.01)
        # self.lower_percent_spin.setValue(0.05)
        # # self.lower_percent_spin.valueChanged.connect(self.upper_or_lower_limit_changed)
        # self.lower_percent_spin.valueChanged.connect(lambda x: delayed_preview_update.start(600))
        #
        # color_coding_layout.addWidget(self.lower_percent_spin)
        # color_coding_layout.addWidget(self.upper_percent_spin)
        #
        # lut_label = QLabel('Lut: ')
        # color_coding_layout.addWidget(lut_label)
        #
        # # populate LUTs --> once for good at the beginning of the table
        # self.lut_combo = QComboBox()
        # available_luts = list_availbale_luts()
        # for lut in available_luts:
        #     self.lut_combo.addItem(lut)
        #
        # self.lut_combo.currentTextChanged.connect(self.lut_changed)
        # color_coding_layout.addWidget(self.lut_combo)
        #
        # self.groupBox_color_coding.setLayout(color_coding_layout)
        #
        # # content_widget.layout.addLayout(color_coding_layout)
        # content_widget.layout.addWidget(self.groupBox_color_coding)
        #
        # self.groupBox_overlay = QGroupBox('Overlay/Blend')
        # self.groupBox_overlay.setToolTip("Overlay selected image over the selected background image")
        # self.groupBox_overlay.setCheckable(True)
        # self.groupBox_overlay.setChecked(False)
        # self.groupBox_overlay.toggled.connect(self.preview_changed)
        #
        # overlay_layout = QHBoxLayout()
        # overlay_layout.setContentsMargins(0, 0, 0, 0)  # self.groupBox_overlay.
        # overlay_layout.setAlignment(Qt.AlignLeft)
        # # TODO --> maybe show and or hide the transparency settings depending on the selection
        # # self.overlay_check = QCheckBox("Enable overlay/blend")
        # # self.overlay_check.stateChanged.connect(self.preview_changed)
        # # overlay_layout.addWidget(self.overlay_check, 0, 0)
        #
        # overlay_bg_label = QLabel("Background image channel:")  # for overlay
        # overlay_layout.addWidget(overlay_bg_label)
        #
        # self.overlay_bg_channel_combo = QComboBox()
        # overlay_layout.addWidget(self.overlay_bg_channel_combo)
        #
        # overlay_fg_transparency_label = QLabel("Foreground opacity:")
        # overlay_layout.addWidget(overlay_fg_transparency_label)
        #
        # self.overlay_fg_transparency_spin = QDoubleSpinBox()
        # self.overlay_fg_transparency_spin.setSingleStep(0.05)
        # self.overlay_fg_transparency_spin.setRange(0.05, 1.)
        # # self.overlay_fg_transparency_spin.valueChanged.connect(self.preview_changed) # direct update upon value change --> not always very smart... better wait some time
        # self.overlay_fg_transparency_spin.valueChanged.connect(lambda x: delayed_preview_update.start(600))
        #
        # self.overlay_fg_transparency_spin.setValue(0.3)
        # overlay_layout.addWidget(self.overlay_fg_transparency_spin)
        #
        # self.groupBox_overlay.setLayout(overlay_layout)
        # content_widget.layout.addWidget(self.groupBox_overlay)
        # # self.tab3.layout.addLayout(overlay_layout)
        #
        # self.groupBox_export = QGroupBox('Export')
        # export_layout = QHBoxLayout()
        # export_layout.setContentsMargins(0, 0, 0, 0)
        # export_layout.setAlignment(Qt.AlignLeft)
        #
        # self.export_image_button = QPushButton("Single image")
        # self.export_image_button.setToolTip("Export as tif the current view")
        # self.export_image_button.clicked.connect(self.export_image)
        # export_layout.addWidget(self.export_image_button)
        #
        # self.export_stack_button = QPushButton("Stack")
        # self.export_stack_button.setToolTip("Export as a tif stack the current view")
        # self.export_stack_button.clicked.connect(self.export_stack)
        # export_layout.addWidget(self.export_stack_button)
        #
        # self.groupBox_export.setLayout(export_layout)
        # content_widget.layout.addWidget(self.groupBox_export)
        #
        # # TODO --> maybe offer overlays/ image composition here !!! --> in a way very easy but there will be issues sometimes maybe --> just think about it ???
        # # for method in EZDeepLearning.available_model_architectures:
        # #     self.image_preview_combo.addItem(method)
        # content_widget.setLayout(content_widget.layout)
        #
        # self.tab4.layout = QVBoxLayout()
        # self.force_CPU_check = QCheckBox(
        #     'Use CPU for deep learning (slower but often offers more memory) (MUST BE SET IMMEDIATELY AFTER LAUNCHING THE SOFTWARE)')
        # self.force_CPU_check.setToolTip(
        #     "Do not use a graphic card even when a compatible one is available on the system")
        # self.force_CPU_check.stateChanged.connect(self.force_CPU)  # need a fix here
        # self.tab4.layout.addWidget(self.force_CPU_check)
        # # TODO --> do the code to force
        # self.tab4.setLayout(self.tab4.layout)

        self.table_widget = QWidget()
        table_widget_layout = QGridLayout()

        # can maybe ask the nb of threads to use here too
        self.logger_console = QTextBrowser(self)
        self.logger_console.setReadOnly(True)
        self.logger_console.textCursor().movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
        self.logger_console.setHtml('<html>')
        self.logger_console.ensureCursorVisible()
        self.logger_console.document().setMaximumBlockCount(1000) # limits to a 1000 entrees

        if not DEBUG:
            try:
                XStream.stdout().messageWritten.connect(self.set_html_black)
                XStream.stderr().messageWritten.connect(self.set_html_red)
                self.handler = QtHandler()
                self.handler.setFormatter(logging.Formatter(TA_logger.default_format))
                # change default handler for logging
                TA_logger.setHandler(self.handler)
            except:
                traceback.print_exc()

        table_widget_layout.addWidget(self.tabs, 0, 0)
        # ok just store console in a qgroup stuff

        self.groupBox_logging = QGroupBox('Log')
        self.groupBox_logging.setToolTip("Shows current status.")
        self.groupBox_logging.setMinimumWidth(250)
        self.groupBox_logging.setEnabled(True)
        # groupBox layout
        self.groupBox_pretrain_layout = QGridLayout()
        self.groupBox_pretrain_layout.addWidget(self.logger_console, 0, 0)
        self.groupBox_logging.setLayout(self.groupBox_pretrain_layout)

        table_widget_layout.addWidget(self.groupBox_logging, 0,
                                      1)  # not bad but maybe put it somewhere else --> such as on the side of the tabs in a gridlayout with all the necessary stuff!!
        self.table_widget.setLayout(table_widget_layout)

        # self.table_widget.setLayout(table_widget_layout)

        # print('bob')
        # Add tabs to widget
        # table_widget_layout.addWidget(self.tabs)
        # table_widget_layout.addWidget(self.logger_console) # not bad but maybe put it somewhere else --> such as on the side of the tabs in a gridlayout with all the necessary stuff!!
        #
        # table_widget_layout.addWidget(self.tabs, 0, 0)
        # # ok just store console in a qgroup stuff
        #
        # self.groupBox_logging = QGroupBox('Log')
        # self.groupBox_logging.setToolTip("Shows TA current status, method progress, warnings and errors")
        # self.groupBox_logging.setMinimumWidth(250)
        # self.groupBox_logging.setEnabled(True)
        # # groupBox layout
        # self.groupBox_pretrain_layout = QGridLayout()
        # self.groupBox_pretrain_layout.addWidget(self.logger_console, 0, 0)
        # self.groupBox_logging.setLayout(self.groupBox_pretrain_layout)
        #
        # table_widget_layout.addWidget(self.groupBox_logging, 0,
        #                               1)  # not bad but maybe put it somewhere else --> such as on the side of the tabs in a gridlayout with all the necessary stuff!!
        # self.table_widget.setLayout(table_widget_layout)

        # voilà le stackwidget --> et comment on ajoute des trucs dedans --> permet de ne montrer que certains widgets
        # print('bob')
        self.volumeViewer = QWidget()  # maybe get it to simply pop out ???
        # ça c'est juste la layout du stack --> on s'en fout

        # TODO replace that with napari
        # self.volumeViewerUI()
        self.Stack.addWidget(self.paint)
        # self.Stack.addWidget(self.properties_table)
        # self.Stack.addWidget(self.volumeViewer)

        # create a grid that will contain all the GUI interface
        self.grid = QGridLayout()
        # grid.setSpacing(10)

        # grid.addWidget(self.scrollArea, 0, 0)
        self.grid.addWidget(self.Stack, 0, 0)
        # self.grid.addWidget(self.Stack, 0, 0,2,1)
        # self.grid.addWidget(self.list, 0, 1)
        # self.grid.addWidget(self.logger_console, 0, 1)
        # self.grid.addWidget(self.list, 1, 1)
        # The first parameter of the rowStretch method is the row number, the second is the stretch factor. So you need two calls to rowStretch, like this: --> below the first row is occupying 80% and the second 20%
        self.grid.setRowStretch(0, 75)
        self.grid.setRowStretch(2, 25)

        # first col 75% second col 25% of total width
        self.grid.setColumnStretch(0, 75)
        self.grid.setColumnStretch(1, 25)

        # void QGridLayout::addLayout(QLayout * layout, int row, int column, int rowSpan, int columnSpan, Qt::Alignment alignment = 0)
        self.grid.addWidget(self.table_widget, 2, 0, 1, 2)  # spans over one row and 2 columns

        # self.setCentralWidget(self.scrollArea)
        self.setCentralWidget(QFrame())
        self.centralWidget().setLayout(self.grid)

        # self.statusBar().showMessage('Ready')
        statusBar = self.statusBar()  # sets an empty status bar --> then can add messages in it
        self.paint.statusBar = statusBar

        # add progress bar to status bar
        self.pbar = QProgressBar(self)
        self.pbar.setToolTip('Shows current progress')
        self.pbar.setGeometry(200, 80, 250, 20)
        statusBar.addWidget(self.pbar)

        self.stop_all_threads_button = QPushButton('Stop')
        self.stop_all_threads_button.setToolTip(
            "Stops the running function or deep learning prediction as soon as possible.")
        self.stop_all_threads_button.clicked.connect(self.stop_threads_immediately)
        statusBar.addWidget(self.stop_all_threads_button)

        self.about = QPushButton()
        self.about.setIcon(qta.icon('mdi.information-variant', options=[{'scale_factor': 1.5}]))
        self.about.clicked.connect(self.about_dialog)
        self.about.setToolTip('About...')
        statusBar.addWidget(self.about)

        # statusBar.addWidget(self.last_opened_file_name)

        # Set up menu bar
        # self.menuBar = QtWidgets.QMenuBar(self)
        # self.mainMenu = QtWidgets.QMenuBar(self)
        # self.menuBar.setGeometry(QtCore.QRect(0, 0, 664, 20))
        # self.menuBar.setObjectName("menuBar")
        # self.setMenuBar(self.mainMenu)
        # self.setMenuBar(self.menuBar)

        # the esc key removes the selection --> can be useful when the selection takes the whole screen
        escShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Escape), self)
        escShortcut.activated.connect(self.paint.EZFIG_panel.remove_sel)
        escShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        # Setup hotkeys for whole system
        # Delete selected vectorial objects
        # deleteShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Delete), self)
        # deleteShortcut.activated.connect(self.down)
        # deleteShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # # shall I remove this useless shortcut
        # # set drawing window fullscreen
        # fullScreenShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_F), self)
        # fullScreenShortcut.activated.connect(self.fullScreen)
        # fullScreenShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # # set drawing window fullscreen
        # fullScreenShortcut2 = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_F12), self)
        # fullScreenShortcut2.activated.connect(self.fullScreen)
        # fullScreenShortcut2.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # # exit from full screen TODO add quit the app too ??
        # escapeShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Escape), self)
        # escapeShortcut.activated.connect(self.escape)
        # escapeShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # spaceShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Space), self)
        # spaceShortcut.activated.connect(self.nextFrame)
        # spaceShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active
        #
        # backspaceShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Backspace), self)
        # backspaceShortcut.activated.connect(self.prevFrame)
        # backspaceShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        self.blinker = Blinker()
        self.to_blink_after_worker_execution = None
        self.threading_enabled = True

        # self.threads = [] # contains all threads so that I can stop them immediately
        # self.thread = None
        # self.event_stop = threading.Event() # allows to stop a thread
        # self.threadpool = QThreadPool()
        # self.threadpool.setMaxThreadCount(self.threadpool.maxThreadCount() - 1)  # spare one core for system

        self.overlay = Overlay(self.centralWidget())  # maybe ok if draws please wait somewhere
        self.overlay.hide()

        # MainWindow.setCentralWidget(self.centralWidget)
        self.mainToolBar = QtWidgets.QToolBar(self)
        self.mainToolBar.setObjectName("mainToolBar")
        self.addToolBar(QtCore.Qt.TopToolBarArea, self.mainToolBar)
        self.statusBar = QtWidgets.QStatusBar(self)
        self.statusBar.setObjectName("statusBar")
        self.setStatusBar(self.statusBar)
        self.menuBar = QtWidgets.QMenuBar(self)
        # self.menuBar.setGeometry(QtCore.QRect(0, 0, 664, 20))
        # self.menuBar.setObjectName("menuBar")
        self.menuMenu = QtWidgets.QMenu(self.menuBar)
        self.menuMenu.setObjectName("menuMenu")
        self.setMenuBar(self.menuBar)
        # self.menuBar.addAction(self.menuMenu.menuAction())
        file_menu = self.menuBar.addMenu("File")
        export_as_svg_action = QAction("Export as svg", self)
        file_menu.addAction(export_as_svg_action)
        export_as_svg_action.triggered.connect(self.export_as_svg)

        export_as_tif_action = QAction("Export as... (jpg, tif, png)", self)
        file_menu.addAction(export_as_tif_action)
        export_as_tif_action.triggered.connect(self.export_as_tif)

    def onTabChange(self):
        # TODO --> maybe select row images or columns depending on the selected tab
        # --> always store all the possibilities when a click is made --> TODO
        # python serialize a numpy array --> to serialize images or gather the images --> have sort of embed --> maybe put a warning depending on the stuff
        # maybe various options for aggregation of data and images

        selected_tab_idx = self.tabs.currentIndex()
        selected_tab_name = self.tabs.tabText(selected_tab_idx).lower()

        print(selected_tab_idx, selected_tab_name)
        print('TODO')

        # pass
    def set_html_red(self, text):
        # quick n dirty log coloring --> improve when I have time
        textCursor = self.logger_console.textCursor()
        textCursor.movePosition(QTextCursor.End)
        self.logger_console.setTextCursor(textCursor)
        format = QTextCharFormat()
        format.setForeground(QColor(255, 0, 0))  # red
        self.logger_console.setCurrentCharFormat(format)
        self.logger_console.insertPlainText(text)
        self.logger_console.verticalScrollBar().setValue(self.logger_console.verticalScrollBar().maximum())

    def set_html_black(self, text):
        # quick n dirty log coloring --> improve when I have time
        textCursor = self.logger_console.textCursor()
        textCursor.movePosition(QTextCursor.End)
        self.logger_console.setTextCursor(textCursor)
        format = QTextCharFormat()
        format.setForeground(QColor(0, 0, 0))  # black
        self.logger_console.setCurrentCharFormat(format)
        self.logger_console.insertPlainText(text)

    def stop_threads_immediately(self):
        print('TODO stop_threads_immediately')
        pass

    def about_dialog(self):
        print('TODO about_dialog')
        pass

    def export_as_svg(self):
        # just for debug
        # print('TADA')
        # pass
        # TODO launch the export of the image as a svg file
        # --> call the stuff
        # qsdqsdqsd
        output_file = saveFileDialog(parent_window=self, extensions="Supported Files (*.svg);;All Files (*)",default_ext='.svg')
        self._save(output_file)

    # could do export as raster and handle the stuff
    def export_as_tif(self):
        # just for debug
        # print('TADA')
        # pass
        # TODO launch the export of the image as a svg file
        # --> call the stuff
        # qsdqsdqsd
        # TODO can I save as pdf ???? --> maybe
        output_file = saveFileDialog(parent_window=self, extensions="Supported Files (*.tif *.jpg *.png);;All Files (*)",default_ext='.tif')
        self._save(output_file)

    def _save(self, output_file):
        if output_file is not None:
            self.paint.EZFIG_panel.save(output_file)


if __name__ == '__main__':
    # vraiment pas mal --> j'y suis presque...
    import sys
    app = QApplication(sys.argv)
    w = EZFIG_GUI()
    demo = True
    if demo:
        shapes_to_draw = []
        shapes_to_draw.append(Polygon2D(0, 0, 10, 0, 10, 20, 0, 20, 0, 0, color=0x00FF00))
        shapes_to_draw.append(
            Polygon2D(100, 100, 110, 100, 110, 120, 10, 120, 100, 100, color=0x0000FF, fill_color=0x00FFFF,
                      stroke=2))
        shapes_to_draw.append(Line2D(0, 0, 110, 100, color=0xFF0000, stroke=3))
        shapes_to_draw.append(Rect2D(200, 150, 250, 100, stroke=10, fill_color=0xFF0000))
        shapes_to_draw.append(Ellipse2D(0, 50, 600, 200, stroke=3))
        shapes_to_draw.append(Circle2D(150, 300, 30, color=0xFF0000, fill_color=0x00FFFF))
        shapes_to_draw.append(PolyLine2D(10, 10, 20, 10, 20, 30, 40, 30, color=0xFF0000, stroke=2))
        shapes_to_draw.append(PolyLine2D(10, 10, 20, 10, 20, 30, 40, 30, color=0xFF0000, stroke=2))
        shapes_to_draw.append(Point2D(128, 128, color=0xFF0000, fill_color=0x00FFFF, stroke=0.65))
        shapes_to_draw.append(Point2D(128, 128, color=0x00FF00, stroke=0.65))
        shapes_to_draw.append(Point2D(10, 10, color=0x000000, fill_color=0x00FFFF, stroke=3))
        shapes_to_draw.append(Rect2D(0, 0, 512, 512, color=0xFF00FF, stroke=6))
        img0 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/00.png')
        inset = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
        inset2 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
        inset3 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
        scale_bar = ScaleBar(30, '<font color="#FF00FF">10µm</font>')
        scale_bar.set_P1(0, 0)
        img0.add_object(scale_bar, Image2D.TOP_LEFT)
        img0.add_object(inset3, Image2D.TOP_LEFT)
        img0.add_object(inset,Image2D.BOTTOM_RIGHT)  # ça marche meme avec des insets mais faudrait controler la taille des trucs... --> TODO
        img0.add_object(TAText2D(text='<font color="#FF0000">bottom right3</font>'), Image2D.BOTTOM_RIGHT)
        img0.setLettering('<font color="red">A</font>')
        img0.annotation.append(Rect2D(88, 88, 200, 200, stroke=3, color=0xFF00FF))
        img0.annotation.append(Ellipse2D(88, 88, 200, 200, stroke=3, color=0x00FF00))
        img0.annotation.append(Circle2D(33, 33, 200, stroke=3, color=0x0000FF))
        img0.annotation.append(Line2D(33, 33, 88, 88, stroke=3, color=0x0000FF))
        img0.annotation.append(Freehand2D(10, 10, 20, 10, 20, 30, 288, 30, color=0xFFFF00, stroke=3))
        img0.annotation.append(Point2D(128, 128, color=0xFFFF00, stroke=6))
        img1 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
        test_text = '''
        </style></head><body style=" font-family:'Comic Sans MS'; font-size:22pt; font-weight:400; font-style:normal;">
        <p style="color:#00ff00;"><span style=" color:#ff0000;">toto</span><br />tu<span style=" vertical-align:super;">tu</span></p>
        '''
        img1.setLettering(TAText2D(text=test_text))
        # background-color: orange;
        # span div et p donnent la meme chose par contre c'est sur deux lignes
        # display:inline; float:left # to display as the same line .... --> does that work html to svg
        # https://stackoverflow.com/questions/10451445/two-div-blocks-on-same-line --> same line for two divs
        img2 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/02.png')
        # crop is functional again but again a packing error
        img2.crop(left=60)
        img2.crop(right=30)
        img2.crop(bottom=90)
        img2.crop(top=60)
        img3 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/03.png')
        img4 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/04.png')
        img5 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/05.png')
        img6 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/06.png')
        img7 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/07.png')
        img8 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/08.png')
        img8.annotation.append(Rect2D(60, 60, 100, 100, stroke=20, color=0xFF00FF))
        img9 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/09.png')
        img10 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/10.png')
        import numpy as np
        import matplotlib.pyplot as plt
        t = np.arange(0.0, 2.0, 0.01)
        s = 1 + np.sin(2 * np.pi * t)
        fig, ax = plt.subplots()
        ax.plot(t, s)
        ax.set(xlabel='time (s)', ylabel='voltage (mV)',
               title='About as simple as it gets, folks')
        ax.grid()
        # graph2d = VectorGraphics2D(fig)
        # graph2d.crop(all=20)  # not great neither
        # vectorGraphics = VectorGraphics2D('/E/Sample_images/sample_images_svg/cartman.svg')
        # vectorGraphics.crop(left=10, right=30, top=10, bottom=10)
        # animatedVectorGraphics = VectorGraphics2D('/E/Sample_images/sample_images_svg/animated.svg')
        # animatedVectorGraphics.crop(left=30)  # , top=20, bottom=20
        # row1 = Row(img0, img1, img2, graph2d,animatedVectorGraphics)  # , graph2d, animatedVectorGraphics  # , img6, #, nCols=3, nRows=2 #le animated marche mais faut dragger le bord de l'image mais pas mal qd meme
        # col1 = Column(img4, img5, img6, vectorGraphics)  # , vectorGraphics# , img6, img6, nCols=3, nRows=2,

        # all seems to  work except the vector graphics --> handle that later then

        row1 = Row(img0, img1, img2)  # , graph2d, animatedVectorGraphics  # , img6, #, nCols=3, nRows=2 #le animated marche mais faut dragger le bord de l'image mais pas mal qd meme
        col1 = Column(img4, img5, img6)  # , vectorGraphics# , img6, img6, nCols=3, nRows=2,
        col2 = Column(img3, img7, img10)
        col1 /= col2
        row2 = Row(img8, img9)
        row2.setLettering('<font color="#FFFFFF">a</font>')
        row1 /= row2
        col1.setToHeight(512)  # creates big bugs now
        row1.setToWidth(512)
        row1.set_P1(128,128)
        shapes_to_draw.append(col1)
        shapes_to_draw.append(row1)
        img4.setLettering('<font color="#0000FF">Z</font>')  # ça marche mais voir comment faire en fait
        shapes_to_draw.append(Square2D(300, 260, 250, stroke=3))


        # first implement the magic selection going inner an object
        # not bad --> it shows a copy of the rects --> that could be used to do a preview --> finalize that
        # see how complex this would be
        # could do the same with all possible shapes for source and target --> TODO
        # detect all the possibilities
        print(type(row1)) # --> a column --> clone it with rects
        # really need implement cloning anyway
        for iii,rect in enumerate(row1):
            print(rect) # -> ce truc est un rect2D --> just need to clone it
            # print(type(rect), isinstance(rect, Rect2D)) # -- > ok --> seems to work
            rect_copy = Rect2D(rect.boundingRect()) # ça marche --> ça cree une copy du rect # and update the colors when
            rect_copy.fill_color = 0xFF0000
            rect_copy.translate(25,25)
            shapes_to_draw.append(rect_copy)
            # if iii==3:
            break


        # does that work ??? --> is my error an error of scaling ???

        # I need implement the cloning from inside
        # images_bounds = []
        # for rect in col1:
        #     print(rect) # -> ce truc est un rect2D --> just need to clone it
        #     # print(type(rect), isinstance(rect, Rect2D)) # -- > ok --> seems to work
        #     # --> not ok because keeps the original rect dimension, not taking into account the scaling --> just get bounds and duplictae
        #     rect_copy = Rect2D(rect.boundingRect()) # ça marche --> ça cree une copy du rect # and update the colors when
        #     # rect_copy = rect_copy*rect.scaling # --> marche pas --> would need to clone the stuff
        #     rect_copy.fill_color = 0x00FF00
        #     rect_copy.translate(25,25)
        #     shapes_to_draw.append(rect_copy)
            # images_bounds.append(rect_copy)
            # marche pas car scaling non preservé --> voir comment faire en fait
        # cree un clone de la classe mais un peu complexe en fait --> car
        # clone_rect_type = type(col1)
        # new_clone_with_rect = clone_rect_type(*images_bounds)
        # new_clone_with_rect.setX(col1.x())
        # new_clone_with_rect.setY(col1.y())
        # new_clone_with_rect.setWidth(col1.width())
        # new_clone_with_rect.setHeight(col1.height())
        #
        # shapes_to_draw.append(new_clone_with_rect)
        # put a + and add a rect to it maybe
        # see


        w.paint.EZFIG_panel.shapes_to_draw = shapes_to_draw # TODO -> make a function for that and hide it so that it is not possible to access it directly
        w.paint.EZFIG_panel.update_size()





        # ça marche --> tt marche --> voir comment faire la suite

    # w.setWindowIcon(app_icon)

    # NB il y a des bugs d'elements qui sont à droite --> petit pb d'alignement --> probablement un pb
    # faire toujours fitter le plot à sa taille --> TODO
    #

    w.show()
    sys.exit(app.exec_())
